import click

from deploifai.cli.context import (
    pass_deploifai_context_obj,
    DeploifaiContextObj,
    is_authenticated,
)
from deploifai.cli.utilities.config import local_config
from deploifai.cli.api import DeploifaiAPIError
from InquirerPy import prompt
import os


@click.command()
@click.argument("name")
@pass_deploifai_context_obj
@is_authenticated
def create(context: DeploifaiContextObj, name: str):
    """
    Create a new project
    """
    deploifai_api = context.api

    command_workspace = context.global_config["WORKSPACE"]["username"]

    # ensure project name is unique to the chosen workspace
    try:
        fragment = """
                fragment project on Project {
                    id name 
                }
                """
        projects = deploifai_api.get_projects(command_workspace, fragment)
    except DeploifaiAPIError as err:
        click.echo("An error occurred when fetching projects. Please try again.")
        return

    project_names = [project["name"] for project in projects]
    user_pwd_dir_names = os.listdir()

    is_valid_name = True

    err_msg = ""

    if name in user_pwd_dir_names:
        is_valid_name = False
        err_msg = f"There are existing files/directories in your computer also named {name}"
    elif name in project_names:
        is_valid_name = False
        err_msg = f"Project name taken. Existing names in chosen workspace: {' '.join(project_names)}."
    elif not name.isalnum():
        is_valid_name = False
        err_msg = "Project name should only contain alphanumeric characters."

    if not is_valid_name:
        click.secho(err_msg, fg="red")
        raise click.Abort()

    # extract and use the cloud profile information
    try:
        cloud_profiles = deploifai_api.get_cloud_profiles(workspace=command_workspace)
    except DeploifaiAPIError:
        click.echo("Could not fetch cloud profiles. Please try again.")
        return

    if not cloud_profiles:
        click.secho("No cloud profiles found. To create a cloud profile: deploifai cloud-profile create", fg="yellow")
        raise click.Abort()
    choose_cloud_profile = prompt(
        {
            "type": "list",
            "name": "cloud_profile",
            "message": "Choose a cloud profile for project",
            "choices": [
                {
                    "name": "{name}({workspace}) - {provider}".format(
                        name=cloud_profile.name,
                        workspace=cloud_profile.workspace,
                        provider=cloud_profile.provider,
                    ),
                    "value": cloud_profile,
                }
                for cloud_profile in cloud_profiles
            ],
        }
    )

    cloud_profile = choose_cloud_profile["cloud_profile"]

    # create project in the backend
    try:
        project_fragment = """
                fragment project on Project {
                    id
                }
        """
        project_id = deploifai_api.create_project(name, cloud_profile, project_fragment)["id"]
    except DeploifaiAPIError as err:
        click.secho(err, fg="red")
        raise click.Abort()

    click.secho(f"Successfully created new project named {name}.", fg="green")

    # create a project directory locally, along with .deploifai directory within this project
    try:
        os.mkdir(name)
    except OSError:
        click.secho("An error occured when creating the project locally", fg="red")
        raise click.Abort()

    click.secho(f"A new directory named {name} has been created locally.", fg="green")

    project_path = os.path.join(os.getcwd(), name)
    local_config.create_config_files(project_path)

    # storing the project information in the local.cfg file
    context.local_config = local_config.read_config_file()
    # set id in local config file
    local_config.set_project_config(project_id, context.local_config)



import typing
from pathlib import Path

import boto3

from deploifai.cli.api import DeploifaiAPI
from deploifai.cli.clouds.utilities.data_storage.handler import DataStorageHandler


class AWSDataStorageHandler(DataStorageHandler):
    def __init__(self, api: DeploifaiAPI, dataset_id: str):
        data = api.get_data_storage_info(dataset_id)

        container_cloud_name = data["containers"][0]['cloudName']

        aws_config_info = data["cloudProviderYodaConfig"]["awsConfig"]

        client = boto3.resource("s3",
                                region_name=aws_config_info["awsRegion"],
                                aws_access_key_id=aws_config_info["awsAccessKey"],
                                aws_secret_access_key=aws_config_info["awsSecretAccessKey"]
                                )

        super().__init__(dataset_id, container_cloud_name, client)

    @staticmethod
    def upload_file(client, file_path: Path, object_key: str, container_cloud_name: str):
        bucket = client.Bucket(container_cloud_name)
        obj = bucket.Object(object_key)

        with open(str(file_path), 'rb') as data:
            obj.upload_fileobj(data)

    def list_files(self, prefix: str = None) -> typing.Generator:
        if prefix is None:
            return self.client.Bucket(self.container_cloud_name).objects.all()
        return self.client.Bucket(self.container_cloud_name).objects.filter(Prefix=prefix)

    @staticmethod
    def download_file(
            client, file, dataset_directory: Path, container_cloud_name: str
    ):
        # getting the name(key) for each file and creating the folders as required
        key = file.key

        if '/' in key:
            DataStorageHandler.make_dirs(key, dataset_directory)

        file_path = str(dataset_directory.joinpath(key))
        bucket = client.Bucket(container_cloud_name)
        obj = bucket.Object(key)

        with open(file_path, 'wb') as file:
            obj.download_fileobj(file)



import typing
from pathlib import Path

from azure.storage.blob import ContainerClient, BlobServiceClient, BlobProperties

from deploifai.cli.api import DeploifaiAPI
from deploifai.cli.clouds.utilities.data_storage.handler import DataStorageHandler


class AzureDataStorageHandler(DataStorageHandler):
    def __init__(self, api: DeploifaiAPI, dataset_id: str):
        data = api.get_data_storage_info(dataset_id)

        self.storage_account_name = data["cloudProviderYodaConfig"]["azureConfig"]["storageAccount"]
        self.storage_access_key = data["cloudProviderYodaConfig"]["azureConfig"]['storageAccessKey']
        container_cloud_name = data["containers"][0]['cloudName']

        account_url = "{account_name}.blob.core.windows.net".format(
            account_name=self.storage_account_name
        )
        blob_service_client = BlobServiceClient(
            account_url=account_url,
            credential=self.storage_access_key,
        )
        client = blob_service_client.get_container_client(container_cloud_name)

        super().__init__(dataset_id, container_cloud_name, client)

    @staticmethod
    def upload_file(
            client: ContainerClient, file_path: Path, object_key: str, container_cloud_name: str
    ):
        blob_client = client.get_blob_client(object_key)
        blob_bytes = file_path.read_bytes()
        blob_client.upload_blob(
            data=blob_bytes, length=file_path.stat().st_size, overwrite=True
        )

    def list_files(self, prefix: str = None) -> typing.Generator:
        if prefix is None:
            return self.client.list_blobs()
        return self.client.list_blobs(name_starts_with=prefix)

    @staticmethod
    def download_file(
            client: ContainerClient, file: BlobProperties, dataset_directory: Path, container_cloud_name: str
    ):
        file_name = file.name
        blob_client = client.get_blob_client(file_name)
        if "/" in file_name:
            DataStorageHandler.make_dirs(file_name, dataset_directory)

        download_file_path = dataset_directory.joinpath(file_name)
        with open(download_file_path, "wb") as download_file:
            blob_client.download_blob().readinto(download_file)



def parse_resource_stanzas_from_formula(formula):
    """Parse the resource stanzas from the formula."""
    stanzas = []
    not_end = False

    for line in formula:
        s = line.lstrip()
        if s.startswith("resource"):
            stanzas.append(line)
            not_end = True
        elif not_end:
            stanzas[-1] += line
            if s.startswith("end"):
                stanzas[-1] += "\n"
                not_end = False

    return stanzas


def generate_new_formula(placeholder_formula, resource_stanzas):
    """Generate the new formula."""

    new_formula = []

    for line in placeholder_formula:
        if line.lstrip().startswith("## RESOURCES ##"):
            new_formula += resource_stanzas
        else:
            new_formula.append(line)

    return new_formula


if __name__ == "__main__":
    formula_file = open("deploifai.rb", "r")
    resource_stanzas = parse_resource_stanzas_from_formula(formula_file.readlines())
    formula_file.close()
    print(f"Number of resource stanzas: {len(resource_stanzas)}")

    placeholder_formula_file = open("deploifai.placeholder.rb", "r")
    new_formula = generate_new_formula(
        placeholder_formula_file.readlines(), resource_stanzas
    )
    placeholder_formula_file.close()
    print(f"Number of lines in new formula: {len(new_formula)}")

    new_formula_file = open("deploifai.rb", "w")
    new_formula_file.writelines(new_formula)
    new_formula_file.close()
    print("New formula written to deploifai.rb")



import click

from deploifai.cli.context import (
    pass_deploifai_context_obj,
    DeploifaiContextObj,
    is_authenticated, project_found,
)


@click.command()
@pass_deploifai_context_obj
@is_authenticated
@project_found
def info(context: DeploifaiContextObj):
    """
    Get information about a project
    """

    project_id = context.local_config["PROJECT"]["id"]

    fragment = """
                fragment project on Project {
                    name status
                    account{
                        username
                    }
                    cloudProfile{
                        provider
                    }
                    experiments{
                        name status environment
                    }
                    dataStorages{
                        name status
                    }
                    trainings{
                        name status state
                    }
                    applications{
                        name status
                    }
                }
                """
    project_data = context.api.get_project(project_id=project_id, fragment=fragment)

    click.secho("Workspace Name: {}".format(project_data["account"]["username"]), fg="blue")

    click.secho("Project Name: {}".format(project_data["name"]), fg="blue")

    click.secho("Cloud Provider: {}".format(project_data["cloudProfile"]["provider"]))

    click.echo("Experiments:")
    for experiment in project_data["experiments"]:
        click.echo("{} <{}> - {}".format(experiment["name"], experiment["status"], experiment["environment"]))

    click.echo("Datasets:")
    for dataset in project_data["dataStorages"]:
        click.echo("{} <{}>".format(dataset["name"], dataset["status"]))

    click.echo("Training Servers:")
    for server in project_data["trainings"]:
        click.echo("{} <{}> - {}".format(server["name"], server["status"], server["state"]))

    click.echo("Deployments:")
    for application in project_data["applications"]:
        click.echo("{}".format(application["name"]))



import click
from deploifai.cli.context import (
    pass_deploifai_context_obj,
    DeploifaiContextObj,
    is_authenticated,
)
from deploifai.cli.utilities.user import parse_user_profiles


@click.command('set')
@click.argument("workspace")
@pass_deploifai_context_obj
@is_authenticated
def set_workspace(context: DeploifaiContextObj, workspace: str):
    """
    Changes current workspace
    """
    deploifai_api = context.api

    user_data = deploifai_api.get_user()
    personal_workspace, team_workspaces = parse_user_profiles(user_data)

    workspaces_from_api = [personal_workspace] + team_workspaces

    # checking validity of workspace, and prompting workspaces choices if not specified
    if workspace and len(workspace):
        for w in workspaces_from_api:
            if w["username"] == workspace:
                command_workspace = w
                break
        else:
            # the workspace user input does not match with any of the workspaces the user has access to
            click.secho(
                f"{workspace} cannot be found. Please put in a workspace you have access to.",
                fg="red",
            )
            raise click.Abort()

    current_workspace = context.global_config["WORKSPACE"]["username"]

    if current_workspace == command_workspace["username"]:
        click.secho(f"Already in the workspace {current_workspace}", fg="yellow")
        raise click.Abort()

    context.global_config["WORKSPACE"]["username"] = command_workspace["username"]

    context.save_config()

    click.secho(f"Successfully switched to {command_workspace['username']}", fg="green")


import json
import typing
from pathlib import Path

from google.cloud import storage
from google.oauth2.service_account import Credentials

from deploifai.cli.api import DeploifaiAPI
from deploifai.cli.clouds.utilities.data_storage.handler import DataStorageHandler


class GCPDataStorageHandler(DataStorageHandler):
    def __init__(self, api: DeploifaiAPI, dataset_id: str):
        data = api.get_data_storage_info(dataset_id)

        container_cloud_name = data["containers"][0]['cloudName']

        gcp_config_info = data["cloudProviderYodaConfig"]["gcpConfig"]

        service_account_key_json = json.loads(gcp_config_info["gcpServiceAccountKey"])

        credentials = Credentials.from_service_account_info(service_account_key_json)

        client = storage.Client(credentials=credentials)

        super().__init__(dataset_id, container_cloud_name, client)

    @staticmethod
    def upload_file(client, file_path: Path, object_key: str, container_cloud_name: str):
        bucket_client = client.get_bucket(container_cloud_name)
        blob_client = bucket_client.blob(object_key)
        blob_client.upload_from_filename(str(file_path))

    def list_files(self, prefix: str = None) -> typing.Generator:
        if prefix is None:
            return self.client.list_blobs(self.container_cloud_name)
        return self.client.list_blobs(self.container_cloud_name, prefix=prefix)

    @staticmethod
    def download_file(
            client, file, dataset_directory: Path, container_cloud_name: str
    ):
        # getting the name for each file and creating the folders as required
        name = file.name
        if '/' in name:
            DataStorageHandler.make_dirs(name, dataset_directory)

        bucket_client = client.get_bucket(container_cloud_name)
        blob_client = bucket_client.blob(name)

        file_path = str(dataset_directory.joinpath(name))

        blob_client.download_to_filename(file_path)



import click
from deploifai.cli.context import (
    pass_deploifai_context_obj,
    DeploifaiContextObj,
    is_authenticated,
)
from deploifai.cli.utilities.user import parse_user_profiles


@click.command('list')
@pass_deploifai_context_obj
@is_authenticated
def list_workspace(context: DeploifaiContextObj):
    """
    Lists out all available workspaces
    """
    deploifai_api = context.api

    user_data = deploifai_api.get_user()
    personal_workspace, team_workspaces = parse_user_profiles(user_data)

    workspaces_from_api = [personal_workspace] + team_workspaces

    current_workspace = context.global_config["WORKSPACE"]["username"]

    click.secho("All accessible workspaces:", fg="blue")

    for workspace in workspaces_from_api:
        workspace_type = "TEAM" if workspace["isTeam"] else "PERSONAL"
        context.debug_msg(workspace)
        is_current_workspace = "*" if workspace["username"] == current_workspace else " "

        click.echo(f"{is_current_workspace} {workspace['username']} - {workspace_type}")
    return



import click

from deploifai.cli.context import (
    pass_deploifai_context_obj,
    DeploifaiContextObj,
    is_authenticated,
)


@click.command("list")
@pass_deploifai_context_obj
@is_authenticated
def list_project(context: DeploifaiContextObj):
    """
    Command to list all projects in the current workspace
    """

    current_workspace = context.global_config["WORKSPACE"]["username"]

    click.secho("Workspace Name: {}".format(current_workspace), fg="blue")

    fragment = """
                    fragment project on Project {
                        id
                        name
                    }
                    """
    projects_data = context.api.get_projects(workspace=current_workspace, fragment=fragment)

    if len(projects_data) == 0:
        click.secho("No projects exist", fg="yellow")
        return

    # If more information is needed change fragment to add it, then change the final echo to provide that info too

    click.secho("All projects:", fg="blue")
    for projects in projects_data:
        click.echo(projects["name"])
    return



import functools
import os
import click
import configparser
import requests
from click import pass_context

from .api import DeploifaiAPI
from .utilities import environment
from .utilities.config import dataset_config, local_config

from deploifai.cli.utilities.credentials import get_auth_token

APP_NAME = "deploifai-cli"

global_config_filename = "deploifai.cfg"
global_config_directory = (
    click.get_app_dir(APP_NAME)
    if not environment.is_development_env
    else os.path.dirname(os.path.realpath(__file__))
)
global_config_filepath = os.path.join(global_config_directory, global_config_filename)
global_config_sections = ["AUTH", "WORKSPACE"]

debug_levels = ["info", "warning", "error"]


class DeploifaiContextObj:
    global_config = configparser.ConfigParser()
    local_config = configparser.ConfigParser()
    dataset_config = configparser.ConfigParser()
    debug = False
    debug_level = "info"
    api: DeploifaiAPI = None

    def __init__(self):
        pass

    def read_config(self):
        # read the global config file
        self.global_config.read(global_config_filepath)

        # initialise sections if they don't exist already
        for section in global_config_sections:
            if section not in self.global_config.sections():
                self.global_config[section] = {}

        self.debug_msg(f"Read global config file from {global_config_filepath}")

        # read local config file
        self.local_config = local_config.read_config_file()

        # read the dataset config file
        self.dataset_config = dataset_config.read_config_file()

    def save_config(self):
        # create global config directory if it doesn't exist already
        if not os.path.isdir(global_config_directory):
            try:
                os.makedirs(global_config_directory)
                self.debug_msg("Created directory to save global config")
            except OSError:
                self.debug_msg(
                    "Error creating directory to store global config file",
                    level="error",
                )
                return

        with open(global_config_filepath, "w") as configfile:
            self.global_config.write(configfile)
            self.debug_msg(f"Saved global config file as {global_config_filepath}")

    def initialise_api(self):
        if "username" in self.global_config["AUTH"]:
            token = get_auth_token(self.global_config["AUTH"]["username"])
            self.api = DeploifaiAPI(token)
        else:
            self.api = DeploifaiAPI()

    def debug_msg(self, message, level="info", **kwargs):
        if self.debug:
            debug_level_index = debug_levels.index(self.debug_level)
            message_level_index = debug_levels.index(level)

            # only log debug messages if the message level index is higher than the debug level index
            if message_level_index >= debug_level_index:
                fg = "blue"
                if level == "warning":
                    fg = "yellow"
                elif level == "error":
                    fg = "red"

                kwargs.update({"fg": kwargs["fg"] if "fg" in kwargs else fg})

                click.secho(message, **kwargs)

    def is_authenticated(self):
        if "username" not in self.global_config["AUTH"]:
            return False
        username = self.global_config["AUTH"]["username"]

        url = f"{environment.backend_url}/auth/check/cli"
        token = get_auth_token(username)

        response = requests.post(
            url,
            json={"username": username},
            headers={"authorization": token},
        )

        if response.status_code == 200:
            return True
        return False


pass_deploifai_context_obj = click.make_pass_decorator(DeploifaiContextObj, ensure=True)


def is_authenticated(f):
    @pass_context
    def wrapper(click_context, *args, **kwargs):
        deploifai_context = click_context.find_object(DeploifaiContextObj)

        if "username" in deploifai_context.global_config["AUTH"]:
            username = deploifai_context.global_config["AUTH"]["username"]

            url = f"{environment.backend_url}/auth/check/cli"
            token = get_auth_token(username)

            response = requests.post(
                url,
                json={"username": username},
                headers={"authorization": token},
            )

            if response.status_code == 200:
                return click_context.invoke(f, *args, **kwargs)

        click.echo(
            click.style("You are not logged in, to log in: ", fg="red")
            + click.style("deploifai auth login", fg="blue")
        )
        return

    return functools.update_wrapper(wrapper, f)


def project_found(f):
    @pass_context
    def wrapper(click_context, *args, **kwargs):
        deploifai_context = click_context.find_object(DeploifaiContextObj)
        if (
                deploifai_context.local_config is not None
                and "PROJECT" in deploifai_context.local_config
                and "id" in deploifai_context.local_config["PROJECT"]
        ):
            return click_context.invoke(f, *args, **kwargs)

        click.echo("Deploifai project not found, to create a new project:")
        click.secho("deploifai project create <project name>", fg="blue")

    return functools.update_wrapper(wrapper, f)


def dataset_found(f):
    @pass_context
    def wrapper(click_context, *args, **kwargs):
        deploifai_context = click_context.find_object(DeploifaiContextObj)
        if (
                deploifai_context.dataset_config is not None
                and "DATASET" in deploifai_context.dataset_config
                and "id" in deploifai_context.dataset_config["DATASET"]
        ):
            dataset_id = deploifai_context.dataset_config["DATASET"]["id"]
            data = deploifai_context.api.get_data_storage_info(dataset_id)
            if data is not None:
                if data["status"] == "DEPLOY_SUCCESS":
                    return click_context.invoke(f, *args, **kwargs)
                else:
                    click.secho("Cannot use this dataset, its status is {}".format(data["status"]), fg="red")
            else:
                click.secho("dataset.cfg found but id is invalid", fg="red")
                click.secho("Please delete the current dataset.cfg, and run deploifai dataset init again", fg="blue")
        else:
            click.secho("dataset.cfg not found, use deploifai dataset init to initialize a dataset", fg="red")

    return functools.update_wrapper(wrapper, f)



import os
import io
import zipfile
import click
import requests

from deploifai.cli.context import pass_deploifai_context_obj, DeploifaiContextObj

# TODO: figure out how to make this usable
# This clones the deploifai/basic-starter-template which is intended for ML deployments
@click.command()
@pass_deploifai_context_obj
@click.argument("name")
def init(deploifai: DeploifaiContextObj, name: str):
    """
    Initialise a new project called NAME
    """
    url = "https://api.github.com/repos/deploifai/basic-starter-template/zipball/main"
    r = requests.get(url)

    with zipfile.ZipFile(io.BytesIO(r.content)) as zip_ref:
        # create a new project directory
        directory = os.path.join(os.getcwd(), name)
        if not os.path.isdir(directory):
            try:
                os.makedirs(directory)
                deploifai.debug_msg("Created new project directory")
            except OSError:
                deploifai.debug_msg("Error creating project directory", level="error")
                return

        deploifai.debug_msg(f"Zip file namelist: {zip_ref.namelist()}")

        files_list = zip_ref.namelist()[1:]

        for file in files_list:
            filepath = directory + "/" + "/".join(file.split("/")[1:])
            deploifai.debug_msg(f"Write to {filepath}")
            with open(filepath, "wb") as new_file:
                new_file.write(zip_ref.read(file))

        deploifai.debug_msg(f"Extracted zip file contents to {directory}")

    click.echo(f"Created a new project called {name}")



import click
from InquirerPy import prompt

from deploifai.cli.context import pass_deploifai_context_obj, DeploifaiContextObj


@pass_deploifai_context_obj
def change_state(context: DeploifaiContextObj, command: str, ):

    project_id = context.local_config["PROJECT"]["id"]

    fragment = """
                fragment project on Project {
                    account{
                        username
                    }
                    trainings{
                        id name status state
                    }
                }
                """
    project_data = context.api.get_project(project_id=project_id, fragment=fragment)

    click.secho("Workspace Name: {}".format(project_data["account"]["username"]), fg="blue")

    server_info = project_data["trainings"]

    context.debug_msg(server_info)

    choose_server = prompt(
        {
            "type": "list",
            "name": "training server",
            "message": "Choose a training server",
            "choices": [
                {
                    "name": "{} <{}> - {}".format(info["name"], info["status"], info["state"]),
                    "value": info
                }
                for info in server_info
            ],
        }
    )
    if choose_server == {}:
        raise click.Abort()
    server = choose_server["training server"]

    server_id = server["id"]
    context.debug_msg(server_id)
    server_name = server["name"]
    server_status = server["status"]
    server_state = server["state"]

    if command == "start":

        if server_status != "DEPLOY_SUCCESS":
            click.secho("Cannot start training server, since server status is {}.".format(server_status), fg="red")
            raise click.Abort()

        if server_state != "SLEEPING":
            click.secho("Cannot start training server, since server state is {}.".format(server_state), fg="red")
            raise click.Abort()

        server_info = context.api.start_training_server(server_id=server_id)

    elif command == "stop":

        if server_status != "DEPLOY_SUCCESS":
            click.secho("Cannot stop training server, since server status is {}.".format(server_status), fg="red")
            raise click.Abort()

        if server_state != "RUNNING":
            click.secho("Cannot stop training server, since server state is {}.".format(server_state), fg="red")
            raise click.Abort()

        server_info = context.api.stop_training_server(server_id=server_id)

    click.secho("Server Name: {}".format(server_name), fg="blue")

    click.secho("Server State: {}".format(server_info["state"]), fg="blue")


from .environment import frontend_url


def get_dashboard_route(workspace: str) -> str:
    return f"{frontend_url}/dashboard/{workspace}"


def get_project_route(workspace: str, project_name: str) -> str:
    return f"{get_dashboard_route(workspace)}/projects/{project_name}"


def get_dataset_route(workspace: str, project_name: str, dataset_id: str) -> str:
    return f"{get_project_route(workspace, project_name)}/datasets/{dataset_id}"



import os
import click
import re
import docker
from time import sleep
from InquirerPy import prompt
from click_spinner import spinner

from deploifai.cli.api import DeploifaiAPIError
from deploifai.cli.context import pass_deploifai_context_obj, DeploifaiContextObj, is_authenticated, project_found, local_config


@click.command()
@click.argument("name")
@click.option(
    "--env-file",
    help="File path for environment variables (e.g. --env-file path/to/env1 --env-file path/to/env2)",
    multiple=True,
)
@click.option(
    "--env",
    help="Environment variables for the container (e.g. --env VAR1=VAL1 --env VAR2=VAL2)",
    multiple=True,
)
@click.option("--port", "-p", help="Port of container")
@pass_deploifai_context_obj
@is_authenticated
@project_found
def create(context: DeploifaiContextObj, name: str, env_file: tuple, env: tuple, port: int):
    """
    Create or update a new deployment
    """
    # Query required data
    project_id = context.local_config['PROJECT']['id']
    workspace = context.global_config['WORKSPACE']['username']

    # Check if name is taken
    where_application = {
        "name": {"equals": name},
        "project": {"is": {"id": {"equals": project_id}}},
        "status": {"not": {"equals": "DESTROY_SUCCESS"}}
    }
    try:
        applications = context.api.get_applications(workspace=workspace, where_application=where_application)
        if len(applications) > 0:
            click.secho("Application name already exists", fg="red")
            raise click.Abort()
    except DeploifaiAPIError:
        click.secho("Error while querying applications", fg="red")
        raise click.Abort()

    # Check if name is valid
    if len(name) < 4 or len(name) > 30 or not re.match(r"^[a-z0-9-]+$", name):
        click.secho("Application name must contain only lowercase alphanumeric characters, and -, not contain spaces, "
                    "and must be between 4 and 30 characters long", fg="red")
        raise click.Abort()

    # Check if port is valid
    if port:
        try:
            port = int(port)
        except ValueError:
            click.secho("Port must be an integer", fg="red")
            raise click.Abort()
        if port < 0 or port > 65535:
            click.secho("Port must be between 0 and 65535", fg="red")
            raise click.Abort()

    # Check env files
    formatted_env = []
    for file in env_file:
        if not os.path.isfile(file):
            click.secho(f"Environment file does not exist: {file}", fg="red")
            raise click.Abort()
        with open(file, "r") as f:
            env_vars = f.read().splitlines()
            for env_var in env_vars:
                try:
                    key, val = env_var.split("=", maxsplit=1)
                except ValueError:
                    click.secho(f"Invalid environment variable: {env_var} in file {file}\n"
                                f"Environment variables must be in the format of KEY=VALUE", fg="red")
                    raise click.Abort()
                formatted_env.append({"name": key, "value": val})

    # Check env variables
    for env_var in env:
        try:
            key, val = env_var.split("=", maxsplit=1)
        except ValueError:
            click.secho(f"Invalid environment variable: {env_var}\n"
                        f"Environment variables must be in the format of KEY=VALUE", fg="red")
            raise click.Abort()
        formatted_env.append({"name": key, "value": val})

    # Check cloud profile
    existing_cloud_profiles = context.api.get_cloud_profiles(workspace=workspace)
    if len(existing_cloud_profiles) == 0:
        click.secho("Please create a cloud profile with deploifai cloud-profile create", fg="red")
        raise click.Abort()

    cloud_profile = prompt(
        {
            "type": "list",
            "name": "cloud_profile",
            "message": "Choose a cloud profile",
            "choices": [
                {
                    "name": "{name}({workspace}) - {provider}".format(
                        name=profile.name,
                        workspace=profile.workspace,
                        provider=profile.provider,
                    ),
                    "value": profile,
                }
                for profile in existing_cloud_profiles
            ],
        }
    )['cloud_profile']

    # Option for local or remote image
    is_local_image = prompt(
        {
            "type": "confirm",
            "message": "Use local docker image?",
            "default": True,
        }
    )[0]

    if is_local_image:
        # Initialize docker client
        try:
            docker_api = docker.APIClient(base_url='unix://var/run/docker.sock')
        except docker.errors.DockerException:
            click.secho("Docker does not seem to be running on your machine", fg="red")
            raise click.Abort()

        # Choose local image, filter out dangling images
        try:
            local_images = docker_api.images(filters={"dangling": False})
        except docker.errors.APIError:
            click.secho("Error while querying local images", fg="red")
            raise click.Abort()

        if len(local_images) == 0:
            click.secho("No local images found", fg="red")
            raise click.Abort()

        image = prompt(
            {
                "type": "list",
                "name": "image",
                "message": "Choose a local image",
                "choices": [{"name": image['RepoTags'], "value": image} for image in local_images]
            }
        )['image']

        # Prompt for tag
        tag = prompt(
            {
                "type": "input",
                "name": "tag",
                "message": "Image tag (default is 'latest')",
                "default": "latest",
            }
        )['tag']
        tag = tag if tag else "latest"

        # Option for existing or new image repository
        is_new_repository = prompt(
            {
                "type": "list",
                "message": "Use new or existing image repository?",
                "choices": [
                    {"name": "Create new image repository (on your cloud account)", "value": True},
                    {"name": "Use existing image repository (on your cloud account or Docker Hub)", "value": False},
                ]
            }
        )[0]

        if is_new_repository:
            # Create new image repository
            click.echo("Creating new image repository... ", nl=False)
            try:
                with spinner():
                    repository = context.api.create_container_registry(project_id, name, cloud_profile.id)
            except DeploifaiAPIError:
                click.secho("Error while creating new image repository", fg="red")
                raise click.Abort()
            click.echo("Done")
        else:
            # Check if deploifai managed container registries exist
            managed_repositories = context.api.get_container_registries(workspace, cloud_profile.id)
            if len(managed_repositories) == 0:
                click.secho("No existing Deploifai-managed repositories found. External repositories are currently not supported.", fg="red")
                raise click.Abort()
            managed_repository = prompt(
                {
                    "type": "list",
                    "name": "repository",
                    "message": "Choose an existing Deploifai-managed repository",
                    "choices": [{"name": repository['name'], "value": repository} for repository in
                                managed_repositories]
                }
            )['repository']
            repository = context.api.get_container_registry(managed_repository['id'])

        # Tag image
        repository_uri = repository['info']['imageUri']
        docker_api.tag(image=image, repository=repository_uri, tag=tag)
        click.echo(f"Tagged image with tag {tag}")

        # Authenticate docker to repository
        username, password = repository['info']['username'], repository['info']['password']
        click.echo("Authenticating to repository... ", nl=False)
        try:
            with spinner():
                docker_api.login(username=username, password=password, registry=repository['info']['loginServer'])
        except docker.errors.APIError:
            click.secho("Error while authenticating to repository", fg="red")
            raise click.Abort()
        click.echo("Done")

        # Push image
        click.echo(f"Pushing image {repository_uri}:{tag}... ", nl=False)
        try:
            with spinner():
                docker_api.push(repository_uri, tag=tag)
        except docker.errors.APIError:
            click.secho("Error while pushing image", fg="red")
            raise click.Abort()
        click.echo("Done")

        # Use repository uri + tag as image uri
        image_uri = repository_uri + ':' + tag
    else:
        # Use publicly accessible image (assume it is valid)
        image_uri = prompt(
            {
                "type": "input",
                "name": "image_uri",
                "message": "Enter public image name (e.g. nginx)",
            }
        )['image_uri']
        if not image_uri:
            click.secho("Please provide a valid image name", fg="red")
            raise click.Abort()

    # Get port
    if not port:
        port = prompt(
            {
                "type": "number",
                "min_allowed": 0,
                "max_allowed": 65535,
                "default": 80,
                "name": "port",
                "message": "Port of the container image"
            }
        )['port']

    try:
        port = int(port) if port else 80
        if port < 0 or port > 65535:
            raise ValueError
    except ValueError:
        click.secho("Please provide a valid port number", fg="red")
        raise click.Abort()

    # Create plan
    try:
        app_config = context.api.get_application_infrastructure_plans(cloud_provider=cloud_profile.provider)
    except DeploifaiAPIError:
        click.secho("Could not find application infrastructure plans", fg="red")
        raise click.Abort()

    # Get app config
    chosen_config = prompt(
        {
            "type": "list",
            "name": "config",
            "message": f"Choose a size for the application ({cloud_profile.provider})",
            "choices": [
                {
                    "name": f"{config['plan']} ({config['config']})",
                    "value": config,
                }
                for config in app_config
            ],
        }
    )['config']
    config = {'plan': chosen_config['plan']}

    # Create application
    click.echo(f"Creating application with image {image_uri}... ", nl=False)
    try:
        with spinner():
            application = context.api.create_application(project_id, name, cloud_profile.id, config, image_uri, port, formatted_env)
    except DeploifaiAPIError:
        click.secho("\nError while creating application", fg="red")
        raise click.Abort()
    click.secho(f"\nApplication {application['name']} created successfully", fg="green")

    # Poll for deployment status every 10 seconds
    click.echo("Deploying application (this will take a few minutes, and it's safe to CTRL+C) ... ", nl=False)
    with spinner() as s:
        while True:
            app = context.api.get_application(application['id'])
            status = app['status']
            if status == "DEPLOY_SUCCESS":
                s.stop()
                click.secho(f"\nApplication successfully deployed and is up on {app['hostname']}", fg="green")
                break
            elif status == "DEPLOY_ERROR":
                s.stop()
                click.secho("\nDeployment failed. This could be due to invalid or inaccessible docker image."
                            "Please contact us at contact@deploif.ai to troubleshoot.", fg="red")
                break
            else:
                sleep(10)



import configparser
import pathlib
import click

from deploifai.cli.utilities.config.find_config_filepath import find_config_absolute_path

relative_config_filepath = pathlib.Path(".deploifai/local.cfg")

config_file_path = find_config_absolute_path(relative_config_filepath)

"""
Manages a .deploifai/local.cfg file to store configuration info about a project.

The .deploifai directory lives in the root directory of the project. The user should be encouraged to version control this directory.

config dictionary structure:
{
    "PROJECT": {"id": string},
    "DATA_STORAGE": {"id": string}
}
"""

config_sections = ["PROJECT", "DATASET"]


class DeploifaiAlreadyInitialisedError(Exception):
    """
    Exception when Deploifai has already been initialised.
    """

    def __init__(self, message):
        super(DeploifaiAlreadyInitialisedError, self).__init__(message)


class DeploifaiDataAlreadyInitialisedError(Exception):
    """
    Exception when a data directory has been initialised through Deploifai in an ML project.
    """

    def __init__(self, message):
        super(DeploifaiDataAlreadyInitialisedError, self).__init__(message)


class DeploifaiNotInitialisedError(Exception):
    """
    Exception when Deploifai project is not found.
    """

    def __init__(self, message):
        click.echo(click.style("Project not found. To create a project: ", fg="red") +
                   click.style("deploifai project create NAME", fg="blue")
                   )
        super(DeploifaiNotInitialisedError, self).__init__(message)


def create_config_files(new_project_dir: str = None):
    global config_file_path
    """
    Creates the folder .deploifai that stores all the config files.
    :return: None
    """
    if new_project_dir is None:
        local_config_dir = pathlib.Path().joinpath(".deploifai")
    else:
        local_config_dir = pathlib.Path(new_project_dir).joinpath(".deploifai")
    local_config_dir.mkdir(exist_ok=False)

    config_file_path = local_config_dir.joinpath("local.cfg")

    config_file_path.touch(exist_ok=True)

    config = configparser.ConfigParser()

    with config_file_path.open("w") as config_file:
        config.write(config_file)

    click.secho(
        """A .deploifai directory has been created, which contains configuration that Deploifai requires.
            You should version control this directory.""",
        fg="blue",
    )


def read_config_file() -> configparser.ConfigParser:
    """
    Read the config file in the existing .deploifai/local.cfg file
    :return: A ConfigParser that contains all the configs in the config file
    """
    if config_file_path is None:
        return None

    try:
        config = configparser.ConfigParser()
        # read the config file
        config.read(config_file_path)

        for section in config_sections:
            if section not in config.sections():
                config[section] = {}

        return config
    except FileNotFoundError as err:
        click.secho("Missing .deploifai/local.cfg file", fg="red")
        raise err


def save_config_file(config: configparser.ConfigParser):
    """
    Save the local config from the context in the local.cfg file.
    :param config: A ConfigParser object representing config data
    """
    with config_file_path.open("w") as config_file:
        config.write(config_file)


def set_project_config(project_id: str, config: configparser.ConfigParser):
    """
    Set the project config into the existing config file.
    :param project_id: The id of the project from the database.
    :param config: The current config
    :return:
    """
    config["PROJECT"] = {"id": project_id}

    save_config_file(config)


def add_data_storage_config(data_storage_id: str, config: configparser.ConfigParser):
    """
    Add the storage configs in the existing config file. Usually run when the user initialises the directory as a data
    directory.
    :param data_storage_id: The id of the data storage from the database.
    :param config: The current config
    :return:
    """
    if (
            config is not None
            and "DATASET" in config
            and "id" in config["DATASET"]
            and len(config["DATASET"]["id"]) != 0
    ):
        raise DeploifaiDataAlreadyInitialisedError(
            "This project already has a dataset attached to it."
        )

    elif (
            config is not None
            and "DATASET" in config
    ):

        config["DATASET"] = {"id": data_storage_id}

    save_config_file(config)



import click
from InquirerPy import prompt

from deploifai.cli.context import pass_deploifai_context_obj, DeploifaiContextObj, is_authenticated, project_found
from deploifai.cli.utilities.version_comparison import compare, multi_compare
from deploifai.cli.utilities.python_supported import supported

framework_options = ["tensorflow", "pytorch"]
server_options = ["small", "medium", "large"]

server_sizes = {
    "cpu": {"small": "SMALL_CPU", "medium": "MEDIUM_CPU", "large": "LARGE_CPU"},
    "gpu": {"small": "SMALL_GPU", "medium": "MEDIUM_GPU", "large": "LARGE_GPU"},
}


@click.command()
@click.argument("name")
@click.option("--no-dataset", is_flag=True, default=False, show_default=True,
              help="Do not mount any datasets on the server")
@click.option("--dataset", "-d", type=str, help="Name of dataset to be mounted on the server")
@click.option("--cloud-profile", "-c", type=str, help="Name of cloud profile to be used")
@click.option("--gpu/--no-gpu", default=True, show_default=True, help="Create training server with or without GPU")
@click.option("--no-framework", is_flag=True, default=False, show_default=True,
              help="Do not install ML frameworks in the server")
@click.option("--framework", '-fw', type=str,
              help='ML framework to be installed in the server: {}'.format(framework_options))
@click.option("--framework-version", '-fwv', type=str, help="ML framework version to be installed")
@click.option("--python-version", '-py', type=str, help="Python version to be installed (works with selected ML framework)")
@click.option("--server-size", "-size", type=str, help="Size of training server: {}".format(server_options))
@pass_deploifai_context_obj
@is_authenticated
@project_found
def create(context: DeploifaiContextObj, name, no_dataset, dataset, cloud_profile, gpu,
           no_framework, framework, framework_version, python_version, server_size):
    """
    Create a new Training Server
    """
    # checking if provided python version is supported or is substring of a supported version
    if python_version:
        python_supported = supported(python_version)
        if not python_supported:
            click.secho("The python version <{}> provided is not supported by Deploifai".format(python_version), fg="red")
            raise click.Abort()

    # checking if correct framework is provided
    if framework and framework not in framework_options:
        click.secho("The ML framework <{}> provided is not supported by Deploifai".format(framework), fg="red")
        raise click.Abort()

    # checking if framework version exists when framework is not specified
    if framework_version and not framework:
        click.secho("Missing --framework option", fg="red")
        raise click.Abort()

    # validating server size input
    if server_size:
        if server_size not in server_options:
            click.secho("Server size <{}> is not supported. Please choose from one of {}".format(server_size, server_options), fg="red")
            raise click.Abort()

    # assume that the user should be in a project directory, that contains local configuration file
    project_id = context.local_config['PROJECT']['id']
    context.debug_msg(project_id)
    # getting all the information required from the get project api call
    fragment = """
    fragment project on Project {
        name
        account{
            username
        }
        dataStorages{
          id
          name
          cloudProfile{
            provider
          }
        }
    }
    """
    project_data = context.api.get_project(project_id=project_id, fragment=fragment)
    workspace_name = project_data["account"]["username"]

    click.secho("Workspace Name: {}".format(workspace_name), fg="blue")
    click.secho("Project Name: {}".format(project_data["name"]), fg="blue")

    # extracting cloud profile information
    existing_cloud_profiles = context.api.get_cloud_profiles(workspace=workspace_name)
    if len(existing_cloud_profiles) == 0:
        click.secho("Please create a cloud profile with")
        click.secho("deploifai cloud-profile create", fg="yellow")
        raise click.Abort()
    if cloud_profile:
        profile = None
        for existing_cloud_profile in existing_cloud_profiles:
            if existing_cloud_profile.name == cloud_profile:
                profile = existing_cloud_profile
                context.debug_msg("cloud profile info: {}".format(profile))
                break
        if not profile:
            click.secho("Please provide a valid cloud profile name", fg="red")
            raise click.Abort()

    # obtaining the desired dataset
    context.debug_msg(no_dataset)
    if not no_dataset:
        dataset_info = project_data["dataStorages"]
        if dataset:
            pick_dataset = None
            for info in dataset_info:
                if info["name"] == dataset:
                    pick_dataset = info
                    context.debug_msg("dataset_info: {}".format(pick_dataset))
                    break
            if not pick_dataset:
                click.secho("Please provide a valid dataset name", fg="red")
                raise click.Abort()
        else:
            choose_dataset = prompt(
                {
                    "type": "list",
                    "name": "dataset",
                    "message": "Choose a dataset",
                    "choices": [
                        {
                            "name": "{} <{}>".format(info["name"], info["cloudProfile"]["provider"]),
                            "value": info
                        }
                        for info in dataset_info
                    ],
                }
            )
            if choose_dataset == {}:
                raise click.Abort()
            pick_dataset = choose_dataset["dataset"]
        dataset_ids = [pick_dataset["id"]]
    else:
        dataset_ids = None

    # extracting cloud profile information
    if not cloud_profile:
        choose_cloud_profile = prompt(
            {
                "type": "list",
                "name": "cloud_profile",
                "message": "Choose a cloud profile for training server",
                "choices": [
                    {
                        "name": "{name}({workspace}) - {provider}".format(
                            name=profile.name,
                            workspace=profile.workspace,
                            provider=profile.provider,
                        ),
                        "value": profile,
                    }
                    for profile in existing_cloud_profiles
                ],
            }
        )
        if choose_cloud_profile == {}:
            raise click.Abort()
        profile = choose_cloud_profile["cloud_profile"]

    # extracting the cloud profile information
    cloud_profile_id = profile.id
    cloud_profile_provider = profile.provider

    # Obtaining information of python and framework version if using framework

    if not no_framework:
        if not framework:
            questions = [
                {
                    "type": "list",
                    "name": "ML Framework",
                    "message": "Which ML Framework will you be using?",
                    "choices": framework_options
                },
            ]
            version = prompt(questions=questions)
            if version == {}:
                raise click.Abort()
            framework = version["ML Framework"]

        if not python_version:
            if not framework_version:
                variable = {"pythonVersion": {"endsWith": '.0'}}
                distinct = "pythonVersion"
                py_options = context.api.falcon_ml_config_distinct(variable, distinct)
                py_version = [info["pythonVersion"] for info in py_options]
                index, py_option = compare(py_version)
                py_option = [element[:-2] for element in py_option]
                choose_python = prompt(
                    {
                        "type": "list",
                        "name": "python_version",
                        "message": "Choose a version of python you would like to use",
                        "choices": [
                            {
                                "name": "{}".format(option),
                                "value": option,
                            }
                            for option in py_option
                        ],
                    }
                )
                if choose_python == {}:
                    raise click.Abort()
                python_version = choose_python["python_version"]
            elif framework and framework_version:
                variable = {"pythonVersion": {"endsWith": '.0'},
                            "framework": {"equals": framework},
                            "frameworkVersion": {"startsWith": framework_version}}
                distinct = "pythonVersion"
                py_options = context.api.falcon_ml_config_distinct(variable, distinct)
                py_version = [info["pythonVersion"] for info in py_options]
                index, py_option = compare(py_version)
                py_option = [element[:-2] for element in py_option]
                choose_python = prompt(
                    {
                        "type": "list",
                        "name": "python_version",
                        "message": "Choose a version of python you would like to use",
                        "choices": [
                            {
                                "name": "{}".format(option),
                                "value": option,
                            }
                            for option in py_option
                        ],
                    }
                )
                if choose_python == {}:
                    raise click.Abort()
                python_version = choose_python["python_version"]

        if not framework_version:
            variable = {"pythonVersion": {"startsWith": python_version},
                        "framework": {"equals": framework}}
            distinct = "frameworkVersion"
            fm_options = context.api.falcon_ml_config_distinct(variable, distinct)
            fm_version = [info["frameworkVersion"] for info in fm_options]
            index, fm_option = compare(fm_version)
            choose_framework = prompt(
                {
                    "type": "list",
                    "name": "framework_version",
                    "message": "Choose a version of {} you would like to use".format(framework),
                    "choices": [
                        {
                            "name": "{}".format(option),
                            "value": option,
                        }
                        for option in fm_option
                    ],
                }
            )
            if choose_framework == {}:
                raise click.Abort()
            framework_version = choose_framework["framework_version"]

        variable = {"pythonVersion": {"startsWith": python_version},
                    "framework": {"equals": framework},
                    "frameworkVersion": {"startsWith": framework_version}}

        ml_config = context.api.falcon_ml_config(variable)
        if len(ml_config) == 0:
            click.secho("No matches for the versions provided, Please select framework version again",
                        fg="yellow")

            # Selecting framework version
            variable = {"pythonVersion": {"startsWith": python_version},
                        "framework": {"equals": framework}}
            distinct = "frameworkVersion"
            fm_options = context.api.falcon_ml_config_distinct(variable, distinct)
            fm_version = [info["frameworkVersion"] for info in fm_options]
            index, fm_option = compare(fm_version)
            choose_framework = prompt(
                {
                    "type": "list",
                    "name": "framework_version",
                    "message": "Choose a version of {} you would like to use".format(framework),
                    "choices": [
                        {
                            "name": "{}".format(option),
                            "value": option,
                        }
                        for option in fm_option
                    ],
                }
            )
            if choose_framework == {}:
                raise click.Abort()
            framework_version = choose_framework["framework_version"]

            variable = {"pythonVersion": {"startsWith": python_version},
                        "framework": {"equals": framework},
                        "frameworkVersion": {"startsWith": framework_version}}

            ml_config = context.api.falcon_ml_config(variable)

            if len(ml_config) == 0:
                click.secho("No matches for the versions provided, Please check your python and framework version",
                            fg="red")
                raise click.Abort()

        py_version = [info["pythonVersion"] for info in ml_config]
        index, return_list = compare(py_version)
        fm_version = [info["frameworkVersion"] for info in ml_config]
        specific_index = multi_compare(fm_version, index)
        config_info = ml_config[specific_index]

        framework_version = config_info["frameworkVersion"]
        python_version = config_info["pythonVersion"]
        falcon_config_id = config_info["id"]
        click.secho("Python Version: {}".format(python_version), fg="blue")
        click.secho("Framework Version: {}".format(framework_version), fg="blue")

        if gpu:
            cuda_version = config_info["cudaVersion"]
            cudnn_version = config_info["cudnnVersion"]
            click.secho("Cuda Version: {}".format(cuda_version), fg="blue")
            click.secho("Cudnn Version: {}".format(cudnn_version), fg="blue")

    else:
        falcon_config_id = None

    # extracting and choosing server instance size
    falcon_config = context.api.get_training_infrastructure_plans(uses_gpu=gpu, cloud_provider=cloud_profile_provider)
    if server_size:
        if gpu:
            server_family = server_sizes['gpu']
        else:
            server_family = server_sizes["cpu"]
        size = server_family[server_size]

        falcon_config_chosen = None
        for config in falcon_config:
            if config["plan"] == size:
                falcon_config_chosen = config
                break

    else:
        choose_falcon_config = prompt(
            {
                "type": "list",
                "name": "falcon",
                "message": "Choose a server instance size",
                "choices": [
                    {
                        "name": "{plan} ({config_type}) - {provider}".format(
                            plan=config["plan"],
                            config_type=config["config"],
                            provider=cloud_profile_provider,
                        ),
                        "value": config,
                    }
                    for config in falcon_config
                ],
            }
        )
        if choose_falcon_config == {}:
            raise click.Abort()
        falcon_config_chosen = choose_falcon_config["falcon"]
    falcon_plan = falcon_config_chosen["plan"]

    # creating server info and providing its information to user
    server_info = context.api.create_training_server(name=name, data_storage_ids=dataset_ids,
                                                     cloud_profile_id=cloud_profile_id,
                                                     falcon_plan=falcon_plan, uses_gpu=gpu,
                                                     falcon_ml_config_id=falcon_config_id, project_id=project_id)
    server_name = server_info["name"]
    server_status = server_info["status"]

    click.secho("Training Server with name <{}> is being deployed".format(server_name), fg="green")
    context.debug_msg(server_status)



import click

from deploifai.cli.context import pass_deploifai_context_obj, DeploifaiContextObj
from deploifai.cli.utilities import credentials


@click.command()
@pass_deploifai_context_obj
def logout(deploifai: DeploifaiContextObj):
    """
    Logout to remove access
    """
    auth = deploifai.global_config["AUTH"]
    username = auth.get("username")

    if username is None:
        click.secho("Not logged in")
        return

    try:
        credentials.delete_auth_token(username)
        deploifai.debug_msg("Deleted auth token in keyring")

    except Exception as e:
        deploifai.debug_msg(e, level="error")
        click.secho("Logout error")
        return

    auth = deploifai.global_config["AUTH"]
    auth.pop("username")

    workspace = deploifai.global_config["WORKSPACE"]
    workspace.pop("username")

    deploifai.save_config()

    click.secho("Logout success", fg="green")



import click
from .login import login
from .logout import logout
from .status import status


@click.group()
def auth():
    """
    Manage Deploifai's authentication state.
    """
    pass


auth.add_command(login)
auth.add_command(logout)
auth.add_command(status)



from pathlib import Path

import click

from deploifai.cli.clouds.utilities.data_storage.handler import DataStorageHandlerEmptyFilesException
from deploifai.cli.context import DeploifaiContextObj, pass_deploifai_context_obj, is_authenticated, dataset_found
from deploifai.cli.core.data_storage import DataStorage


@click.command()
@click.argument("target", required=False)
@pass_deploifai_context_obj
@is_authenticated
@dataset_found
def pull(context: DeploifaiContextObj, target: str = None):
    """
    Download files from the cloud to local.
    """

    api = context.api
    dataset_id = context.dataset_config["DATASET"]["id"]

    data = context.api.get_data_storage_info(dataset_id)

    click.secho("Dataset Name: {}".format(data["name"]), fg="blue")
    click.secho("Cloud Provider: {}".format(data["cloudProviderYodaConfig"]["provider"]), fg="blue")

    datastorage_handler = DataStorage(api, dataset_id)

    # find the absolute path to the target directory
    cwd = Path.cwd()
    target_abs = cwd if target is None else cwd.joinpath(target)

    try:
        datastorage_handler.pull(target_abs)
    except DataStorageHandlerEmptyFilesException:
        click.secho("No files to pull", fg='yellow')



import click

from .create import create
from .browse import browse
from .list import list_project
from .init import init
from .info import info


@click.group()
def project():
    """
    Initialise, manage, and deploy an ML project
    """


project.add_command(create)
project.add_command(browse)
project.add_command(list_project)
project.add_command(init)
project.add_command(info)



import click
from deploifai.cli.context import (
    pass_deploifai_context_obj,
    DeploifaiContextObj,
    is_authenticated,
)


@click.command("list")
@pass_deploifai_context_obj
@is_authenticated
def list_profile(context: DeploifaiContextObj):
    """
    List all cloud profiles in the current workspace
    """

    current_workspace = context.global_config["WORKSPACE"]["username"]

    click.secho("Workspace Name: {}".format(current_workspace), fg="blue")

    cloud_profile_info = context.api.get_cloud_profiles(current_workspace)

    if len(cloud_profile_info) == 0:
        click.secho("No cloud profiles exist", fg="yellow")
        return

    click.secho("All cloud profiles:", fg="blue")
    for info in cloud_profile_info:
        click.echo(f"{info.name} - {info.provider}")
    return



import click

from deploifai.cli.context import pass_deploifai_context_obj, DeploifaiContextObj, is_authenticated, project_found
from deploifai.cli.utilities.server_state import change_state


@click.command()
@pass_deploifai_context_obj
@is_authenticated
@project_found
def stop(context: DeploifaiContextObj):
    """
        Stop a training server
    """

    change_state("stop")



import click

from .init import init
from .create import create

@click.group()
def application():
    """
    Initialise, and manage an ML application deployment
    """


application.add_command(init)
application.add_command(create)



from pathlib import Path

from deploifai.cli.api import DeploifaiAPI
from deploifai.cli.clouds.azure.data_storage.handler import AzureDataStorageHandler
from deploifai.cli.clouds.aws.data_storage.handler import AWSDataStorageHandler
from deploifai.cli.clouds.gcp.data_storage.handler import GCPDataStorageHandler


class DataStorage:
    def __init__(self, api: DeploifaiAPI, dataset_id: str, handler=None):
        self.api = api
        self.id = dataset_id
        self.handler = handler
        self.register_handler()

    def register_handler(self):
        data = self.api.get_data_storage_info(self.id)
        provider = data["cloudProviderYodaConfig"]["provider"]
        if provider == "AZURE":
            self.handler = AzureDataStorageHandler(self.api, self.id)
        elif provider == "AWS":
            self.handler = AWSDataStorageHandler(self.api, self.id)
        elif provider == "GCP":
            self.handler = GCPDataStorageHandler(self.api, self.id)

    def push(self, target: Path):
        self.handler.push(target)

    def pull(self, target: Path):
        self.handler.pull(target)



import click

from .create import create
from .list import list_profile


@click.group()
def cloud_profile():
    """
    Manage user's cloud profiles
    """


cloud_profile.add_command(create)
cloud_profile.add_command(list_profile)



import click

from deploifai.cli.context import (
    pass_deploifai_context_obj,
    DeploifaiContextObj,
    debug_levels,
)

from .auth import auth
from .application import application
from .project import project
from .dataset import dataset
from .mlflow import mlflow
from .cloud_profile import cloud_profile
from .workspace import workspace
from .server import server

commands = {
    "auth": auth,
    "project": project,
    "dataset": dataset,
    "application": application,
    "cloud-profile": cloud_profile,
    "mlflow": mlflow,
    "workspace": workspace,
    "server": server,
}


@click.group(commands=commands)
@click.version_option(package_name="deploifai", message="%(prog)s %(version)s")
@pass_deploifai_context_obj
@click.option("--debug", is_flag=True, help="Show debug logs")
@click.option(
    "--debug-level",
    type=click.Choice(debug_levels),
    default="info",
    help="Set debugging level",
)
def cli(deploifai: DeploifaiContextObj, debug: bool, debug_level):
    """
    Deploifai CLI
    """
    deploifai.debug = debug
    deploifai.debug_level = debug_level
    deploifai.read_config()
    deploifai.initialise_api()


def main():
    cli(auto_envvar_prefix="DEPLOIFAI")


if __name__ == "__main__":
    main()



import subprocess
import os
import json
import click
import boto3
import logging
from time import sleep
from click_spinner import spinner
from azure.identity import ClientSecretCredential
from azure.core.exceptions import ClientAuthenticationError
from botocore.exceptions import ClientError
from InquirerPy import prompt
from deploifai.cli.utilities.cloud_profile import Provider
from deploifai.cli.context import (
    pass_deploifai_context_obj,
    DeploifaiContextObj,
    is_authenticated,
)
from deploifai.cli.api import DeploifaiAPIError


@click.command()
@click.argument("name")
@click.option(
    "-p",
    "--provider",
    type=click.Choice([p.value for p in Provider]),
    help="Provider for the new cloud profile"
)
@pass_deploifai_context_obj
@is_authenticated
def create(context: DeploifaiContextObj, name: str, provider: str):
    """
    Create a new cloud profile
    """
    deploifai_api = context.api

    command_workspace = context.global_config["WORKSPACE"]["username"]

    try:
        cloud_profiles = deploifai_api.get_cloud_profiles(workspace=command_workspace)
    except DeploifaiAPIError as err:
        click.secho(err, fg="red")
        return

    existing_names = [cloud_profile.name for cloud_profile in cloud_profiles]

    if name in existing_names:
        click.echo(click.style("Cloud profile name taken. Existing names: ", fg="red") + ' '.join(existing_names))
        raise click.Abort()

    if not provider:
        provider = prompt(
            {
                "type": "list",
                "name": "provider",
                "message": "Choose a provider for the new cloud profile",
                "choices": [{"name": p.value, "value": p} for p in Provider]
            }
        )["provider"]

    else:
        provider = Provider(provider)   # Cast to Provider enum

    cloud_credentials = {}
    if provider == Provider.AWS:
        s = boto3.session.Session()
        profiles = s.available_profiles
        profile = 'default'
        if not profiles:
            click.secho('AWS credentials not found. Please set it up with "aws configure", or go to '
                        'https://docs.deploif.ai/cloud-services/connect-your-account/aws to learn how to create AWS '
                        'credentials for Deploifai on the dashboard.', fg="red")
            raise click.Abort()
        if len(profiles) > 1:
            profile = prompt(
                {
                    "type": "list",
                    "name": "profile",
                    "message": "Choose an AWS profile to use",
                    "choices": [{"name": p, "value": p} for p in profiles]
                }
            )["profile"]
            s = boto3.session.Session(profile_name=profile)

        local_creds = s.get_credentials()
        iam = boto3.client('iam', aws_access_key_id=local_creds.access_key, aws_secret_access_key=local_creds.secret_key)

        # Log in attempt - use an arbitrary service
        try:
            iam.list_users()
        except ClientError as err:
            if err.response['Error']['Code'] == 'InvalidClientTokenId':
                click.secho(f"Invalid credentials for profile '{profile}'", fg="red")
            else:
                click.secho(err, fg="red")
            raise click.Abort()
        except Exception as err:
            click.secho(err, fg="red")
            raise click.Abort()

        # Attempt to create user
        try:
            iam.create_user(Path="/", UserName=name)
            click.echo(f"Created IAM user '{name}'")
        except ClientError as err:
            if err.response['Error']['Code'] == 'EntityAlreadyExists':
                click.secho(f"IAM user '{name}' already exists", fg="red")
            else:
                click.secho(err, fg="red")
            raise click.Abort()
        except Exception as err:
            click.secho(err, fg="red")
            raise click.Abort()

        try:
            # Attach policies
            iam.attach_user_policy(UserName=name, PolicyArn="arn:aws:iam::aws:policy/PowerUserAccess")
            iam.attach_user_policy(UserName=name, PolicyArn="arn:aws:iam::aws:policy/IAMFullAccess")
            click.echo("Attached policies")

            # Create access key
            access_key = iam.create_access_key(UserName=name)['AccessKey']
            click.echo("Created access keys")
            cloud_credentials["awsAccessKey"] = access_key['AccessKeyId']
            cloud_credentials["awsSecretAccessKey"] = access_key['SecretAccessKey']
        except Exception as err:
            click.secho(err, fg="red")
            raise click.Abort()
    elif provider == Provider.AZURE:
        # Prompt for account
        res = subprocess.run("az account list -o json", shell=True, capture_output=True)
        if res.returncode != 0:
            click.secho(res.stderr.decode("utf-8").strip(), fg="red")
            raise click.Abort()
        accounts = json.loads(res.stdout.decode('utf-8'))
        if not accounts:
            click.secho("No Azure accounts found. Please log in with 'az login'", fg="red")
            raise click.Abort()
        if len(accounts) == 1:
            account = accounts[0]
        else:
            account = prompt(
                {
                    "type": "list",
                    "name": "account",
                    "message": "Choose an Azure account to use",
                    "choices": [{"name": f"{a['name']} ({a['user']['name']})", "value": a} for a in accounts]
                }
            )["account"]

        # Set azure cli context to the selected account
        subscription_id = account["id"]
        res = subprocess.run(f'az account set --subscription {subscription_id}', shell=True)
        if res.returncode != 0:
            raise click.Abort()
        click.echo(f"Set account to {account['name']} ({account['user']['name']}) ({subscription_id})")

        # Create service principal
        click.echo("Creating new service principal... ", nl=False)
        with spinner():
            res = subprocess.run(f'az ad sp create-for-rbac -n {name} --role Contributor --scopes="/subscriptions/{subscription_id}" -o json', shell=True, capture_output=True)
        if res.returncode != 0:
            click.secho(res.stderr.decode("utf-8").strip(), fg="red")
            raise click.Abort()
        azure_credentials = json.loads(res.stdout.decode('utf-8').strip())
        click.secho(f" \nCreated service principal {name} ({azure_credentials['appId']})", fg="green")  # Note: added space before newline to prevent spinner residue

        tenant_id = azure_credentials['tenant']
        client_id = azure_credentials['appId']
        client_secret = azure_credentials['password']

        cloud_credentials["azureSubscriptionId"] = subscription_id
        cloud_credentials["azureTenantId"] = tenant_id
        cloud_credentials["azureClientId"] = client_id
        cloud_credentials["azureClientSecret"] = client_secret

        click.echo("Creating cloud profile... ", nl=False)

        # Ping azure with newly created credentials until it is ready
        logger = logging.getLogger("azure.identity._internal.get_token_mixin")
        logger_level = logger.getEffectiveLevel()
        logger.setLevel(logging.ERROR)  # temporarily disable noisy azure logs
        with spinner():
            # Exponential backoff (double sleep time on each error)
            sleep_time = 2
            while True:
                # Escape from infinite loop
                if sleep_time > 32:
                    click.secho("Timeout while attempting to create cloud profile with the new service principal "
                                "(this is likely an issue with Azure)", fg="red")
                    raise click.Abort()
                sleep(sleep_time)
                try:
                    # Attempt to create resource group if it does not exist
                    ClientSecretCredential(tenant_id, client_id, client_secret).get_token("https://management.azure.com/.default")
                    break
                except ClientAuthenticationError as e:
                    context.debug_msg(e)
                    sleep_time *= 2
                except Exception as e:
                    context.debug_msg(e)
                    click.secho("Something went wrong with the newly generated service principal credentials", fg="red")
                    raise click.Abort()
        click.echo(" ")  # print space to prevent spinner residue
        logger.setLevel(logger_level)  # restore logger verbosity
    else:
        # Select projects
        res = subprocess.run('gcloud projects list --format=json', shell=True, capture_output=True)
        if res.returncode != 0:
            raise click.Abort()
        gcp_projects = json.loads(res.stdout.decode('utf-8'))

        # Filter out inactive projects
        gcp_projects = [p for p in gcp_projects if p['lifecycleState'] == 'ACTIVE']
        if not gcp_projects:
            click.secho('No active Google Cloud projects found. Please create one using gcloud cli or at https://console.cloud.google.com/home/dashboard', fg="red")
            raise click.Abort()

        gcp_project_id = prompt(
            {
                "type": "list",
                "name": "project",
                "message": "Choose a Google Cloud project to use for this cloud profile",
                "choices": [{"name": f"{p['name']} ({p['projectId']})", "value": p['projectId']} for p in gcp_projects]
            }
        )["project"]

        cloud_credentials["gcpProjectId"] = gcp_project_id

        # Set project
        res = subprocess.run(f'gcloud config set project {gcp_project_id}', shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        if res.returncode != 0:
            raise click.Abort()
        click.echo(f"Set GCP project to {gcp_project_id}")

        # Enable services
        res = subprocess.run('gcloud services enable artifactregistry.googleapis.com compute.googleapis.com iam.googleapis.com iamcredentials.googleapis.com storage-api.googleapis.com', shell=True)
        if res.returncode != 0:
            raise click.Abort()

        # Create service account
        res = subprocess.run(f'gcloud iam service-accounts create {name} --display-name {name} --description "Service account for Deploifai"', shell=True)
        if res.returncode != 0:
            raise click.Abort()

        # Get service account email
        res = subprocess.run(f'gcloud iam service-accounts list --filter="displayName:{name}" --format="value(email)"', shell=True, capture_output=True)
        if res.returncode != 0:
            raise click.Abort()
        service_account_email = res.stdout.decode('utf-8').strip()

        # Attach policies
        roles = ['roles/editor', 'roles/resourcemanager.projectIamAdmin', 'roles/storage.admin', 'roles/run.admin']
        for role in roles:
            res = subprocess.run(f"gcloud projects add-iam-policy-binding {gcp_project_id} --member=serviceAccount:{service_account_email} --role={role}", shell=True, capture_output=True)
            if res.returncode != 0:
                click.secho(res.stderr.decode('utf-8'), fg="red")
                raise click.Abort()
            click.echo(f"Attached role {role}")

        # Generate key file '.key.json' in current directory
        res = subprocess.run(f'gcloud iam service-accounts keys create .key.json --iam-account={service_account_email} --key-file-type=json', shell=True, capture_output=True)
        if res.returncode != 0:
            click.secho(res.stderr.decode('utf-8'), fg="red")
            raise click.Abort()

        # Extract key and delete key file
        try:
            with open(".key.json") as gcp_service_account_key_json:
                cloud_credentials["gcpServiceAccountKey"] = gcp_service_account_key_json.read()
            os.remove(".key.json")
        except FileNotFoundError:
            click.secho("File not found. Please input the correct file path.", fg="red")
            raise click.Abort()

    context.debug_msg(cloud_credentials)

    # Create cloud profile
    cloud_profile_fragment = """
    fragment cloud_profile on CloudProfile {
        id
    }
    """
    # If Azure, retry 5 times
    if provider == Provider.AZURE:
        for i in range(5):
            try:
                _ = deploifai_api.create_cloud_profile(
                    provider,
                    name,
                    cloud_credentials,
                    command_workspace,
                    cloud_profile_fragment
                )
                break
            except DeploifaiAPIError as err:
                context.debug_msg(err)
                if i == 4:
                    click.secho(err, fg="red")
                    raise click.Abort()
                context.debug_msg("Retrying in 5 seconds...")
                sleep(5)
    else:
        try:
            _ = deploifai_api.create_cloud_profile(
                provider,
                name,
                cloud_credentials,
                command_workspace,
                cloud_profile_fragment
            )
        except DeploifaiAPIError as err:
            click.secho(err, fg="red")
            raise click.Abort()

    click.secho(f"Successfully created a new cloud profile: {name}.", fg="green")



import click
import click_spinner
from InquirerPy import prompt
import os
from time import sleep

from deploifai.cli.api import DeploifaiAPIError
from deploifai.cli.context import (
    pass_deploifai_context_obj,
    DeploifaiContextObj,
    is_authenticated,
    project_found,
)
from deploifai.cli.utilities.config import dataset_config


@click.command()
@pass_deploifai_context_obj
@project_found
@is_authenticated
def create(context: DeploifaiContextObj):
    """
    Create a new dataset
    """

    deploifai_api = context.api
    # assume that the user should be in a project directory, that contains local configuration file
    project_id = context.local_config["PROJECT"]["id"]

    # query for workspace name from api
    fragment = """
                    fragment project on Project {
                        name
                        account{
                            username
                        }
                    }
                    """
    context.debug_msg(project_id)
    project_data = context.api.get_project(project_id=project_id, fragment=fragment)
    project_name = project_data["name"]
    command_workspace = project_data["account"]["username"]
    context.debug_msg(command_workspace)
    click.secho("Workspace:{}\n".format(command_workspace), fg='green')
    click.secho("Project:{}<{}>\n".format(project_name, project_id), fg='green')

    try:
        cloud_profiles = deploifai_api.get_cloud_profiles(
            workspace=command_workspace
        )
    except DeploifaiAPIError as err:
        click.echo("Could not fetch cloud profiles. Please try again.")
        return
    new_storage_questions = [
        {
            "type": "input",
            "name": "storage_name_input",
            "message": "Name the dataset",
        },
        {
            "type": "list",
            "name": "cloud_profile",
            "message": "Choose a cloud profile for dataset",
            "choices": [
                {
                    "name": "{name}({workspace}) - {provider}".format(
                        name=cloud_profile.name,
                        workspace=cloud_profile.workspace,
                        provider=cloud_profile.provider,
                    ),
                    "value": cloud_profile,
                }
                for cloud_profile in cloud_profiles
            ],
        },
    ]
    new_storage_answers = prompt(questions=new_storage_questions)
    if new_storage_answers == {}:
        raise click.Abort()
    storage_name = new_storage_answers["storage_name_input"]
    cloud_profile = new_storage_answers["cloud_profile"]

    click.secho("Cloud Profile:{}".format(cloud_profile), fg='green')
    click.secho("Dataset Name:{}".format(storage_name), fg='green')

    click.secho("Creating a new dataset.", fg="blue")

    context.debug_msg(cloud_profile)

    try:
        create_storage_response = deploifai_api.create_data_storage(
            storage_name, project_id, cloud_profile
        )
    except DeploifaiAPIError as err:
        click.secho(err, fg='red')
        raise click.Abort()

    context.debug_msg(create_storage_response)

    with click_spinner.spinner():
        click.echo("Deploying dataset")
        while True:
            data_storage_info = deploifai_api.get_data_storage_info(
                create_storage_response["id"]
            )
            if data_storage_info["status"] == "DEPLOY_SUCCESS":
                click.echo("Deployment success!")
                break
            elif data_storage_info["status"] == "DEPLOY_ERROR":
                click.echo("There was an error in deployment.")
                break
            sleep(10)

    # Obtaining information of cloud provider for dataset created
    storage_id = create_storage_response["id"]
    cloud_data = context.api.get_data_storage_info(storage_id=storage_id)

    dataset_name = cloud_data["name"]
    cloud_provider = cloud_data["cloudProviderYodaConfig"]["provider"]

    click.secho("Storage Name: {}".format(dataset_name), fg="blue")
    click.secho("Cloud Provider: {}".format(cloud_provider), fg="blue")
    context.debug_msg(cloud_data["containers"][0])
    context.debug_msg(cloud_data["cloudProviderYodaConfig"])

    if cloud_provider == "AWS":
        cloud_name = cloud_data["containers"][0]["cloudName"]
        link = "s3://" + cloud_name
        click.secho("AWS s3 link: {}".format(link))
    elif cloud_provider == "GCP":
        cloud_name = cloud_data["containers"][0]["cloudName"]
        link = "gs://" + cloud_name
        click.secho("GCP link: {}".format(link))
    elif cloud_provider == "AZURE":
        cloud_name = cloud_data["containers"][0]["cloudName"]
        azure_account = cloud_data["cloudProviderYodaConfig"]["azureConfig"]["storageAccount"]
        click.secho("account-name: {}".format(azure_account))
        click.secho("container-name: {}".format(cloud_name))

    # create a dataset directory locally, along with dataset.cfg file within
    try:
        os.mkdir(storage_name)
    except OSError:
        click.secho("An error occurred when creating the dataset locally", fg="red")
        click.echo("But the dataset has been created on Deploifai")
        raise click.Abort()

    click.secho(f"A new directory named {storage_name} has been created locally.", fg="green")

    dataset_path = os.path.join(os.getcwd(), storage_name)
    dataset_config.create_config_files(dataset_path)

    # storing the dataset information in the dataset.cfg file
    context.dataset_config = dataset_config.read_config_file()
    # set dataset id in dataset config file
    dataset_config.add_data_storage_config(create_storage_response["id"], context.dataset_config)


def parse_user_profiles(me_user_query):
    personal_workspace = me_user_query["account"]
    team_workspaces = list(map(lambda x: x["account"], me_user_query["teams"]))
    return personal_workspace, team_workspaces



import semver


def compare(input_list):
    given_list = input_list.copy()
    for i in range(len(given_list)):
        for j in range(i + 1, len(given_list)):
            if semver.compare(given_list[i], given_list[j]) == 1:
                temp = given_list[i]
                given_list[i] = given_list[j]
                given_list[j] = temp
    temp = []
    for i, x in enumerate(input_list):
        if x == given_list[-1]:
            temp.append(i)
    return temp, given_list


def multi_compare(version_list, index_list):
    new = {}
    for i in index_list:
        new[i] = version_list[i]
    index = list(new.keys())
    py_value = list(new.values())
    value = py_value.copy()
    for i in range(len(value)):
        for j in range(i + 1, len(value)):
            if semver.compare(value[i], value[j]) == 1:
                temp = value[i]
                value[i] = value[j]
                value[j] = temp
    temp = -1
    for i, x in enumerate(py_value):
        if x == value[-1]:
            temp = i
    index_value = index[temp]
    return index_value



import click
import requests
from deploifai.cli.utilities import environment
from deploifai.cli.utilities import credentials
from deploifai.cli.context import pass_deploifai_context_obj, DeploifaiContextObj


@click.command()
@pass_deploifai_context_obj
@click.option(
    "-u",
    "--username",
    prompt=True,
    help="Username on Deploifai (Optionally can be set via DEPLOIFAI_LOGIN_USERNAME environment variable)",
)
@click.option(
    "-t",
    "--token",
    prompt="Personal access token",
    help="Generated personal access token on Deploifai (Optionally can be set via DEPLOIFAI_LOGIN_TOKEN environment variable)",
)
def login(deploifai: DeploifaiContextObj, username: str, token: str):
    """
    Login using personal access token
    """
    try:
        url = f"{environment.backend_url}/auth/login/cli"
        deploifai.debug_msg(f"Login url: {url}")
        response = requests.post(
            url,
            json={"username": username},
            headers={"authorization": token},
        )
    except requests.exceptions.RequestException as e:
        deploifai.debug_msg(e, level="error")
        click.echo("Error sending network request")
        return

    if response.status_code != 200:
        click.secho("Invalid login", fg="red")
        return

    click.secho("Login success", fg="green")

    deploifai.global_config["AUTH"]["username"] = username
    deploifai.global_config["WORKSPACE"]["username"] = username

    try:
        credentials.save_auth_token(username, token)
        deploifai.debug_msg("Saved auth token in keyring")
    except Exception as e:
        deploifai.debug_msg(e, level="error")
        click.echo("Error saving auth token")
        return

    deploifai.save_config()



from deploifai.cli.api import DeploifaiAPI


def supported(python_version: str):
    variable = {"pythonVersion": {"startsWith": python_version}}
    api = DeploifaiAPI()
    ml_config = api.falcon_ml_config(variable)
    if len(ml_config) == 0:
        python_supported = False
    else:
        python_supported = True

    return python_supported



from enum import Enum

import click
import pyperclip
from InquirerPy import prompt

from deploifai.cli.context import pass_deploifai_context_obj, DeploifaiContextObj, is_authenticated, project_found


class Environment(Enum):
    # Experiment Environment, DEPLOIFAI means usage of a managed runner, EXTERNAL means lack of a managed runner
    DEPLOIFAI = 'DEPLOIFAI'
    EXTERNAL = 'EXTERNAL'


@click.command('get-setup')
@click.option("--external", is_flag=True, help="Use mlflow integration outside of Deploifai managed runner")
@pass_deploifai_context_obj
@is_authenticated
@project_found
def get_setup(context: DeploifaiContextObj, external):
    """
    Get setup to integrate mlflow in training scripts
    """

    # assume that the user should be in a project directory, that contains local configuration file
    project_id = context.local_config['PROJECT']['id']
    # getting all the required information to provide the code
    fragment = """
                fragment project on Project {
                    name
                    account{
                        username
                    }
                    experiments{
                        name
                        environment
                        resourceAccessToken {
                            token
                        }
                    } 
                }
                """
    context.debug_msg(project_id)
    project_data = context.api.get_project(project_id=project_id, fragment=fragment)
    project_name = project_data["name"]
    workspace_name = project_data["account"]["username"]
    experiment_data = project_data["experiments"]

    click.secho("Workspace: {}".format(workspace_name), fg='blue')
    click.secho("Project: {}<{}>".format(project_name, project_id), fg='blue')

    # selecting experiment name from the options
    if len(experiment_data) == 0:
        click.secho("No experiments exist for the project", fg='yellow')
        return
    else:
        choose_experiment = prompt(
            {
                "type": "list",
                "name": "experiment",
                "message": "Choose a experiment",
                "choices": [
                    {
                        "name": "{}".format(experiment_data["name"]),
                        "value": experiment_data
                    }
                    for experiment_data in experiment_data
                ],
            }
        )
        if choose_experiment == {}:
            raise click.Abort()
        experiment = choose_experiment["experiment"]

    experiment_name = experiment["name"]
    experiment_environment = experiment["environment"]
    experiment_token = experiment["resourceAccessToken"]['token']

    click.secho("Experiment: {}".format(experiment_name), fg='blue')

    link = f"{workspace_name}/{project_name}/{experiment_name}"
    line4 = "# setup mlflow for this experiment\n"
    line5 = "mlflow.set_tracking_uri('https://community.mlflow.deploif.ai')\n"
    line6 = 'mlflow.set_experiment("{}")\n'.format(link)

    if external or experiment_environment == Environment.EXTERNAL.value:

        line1 = "# only if running an experiment outside deploifai (this must come before setting mlflow tracking uri and experiment)\n"
        line2 = 'os.environ["MLFLOW_TRACKING_USERNAME"] = "{}" \n'.format(link)
        line3 = 'os.environ["MLFLOW_TRACKING_PASSWORD"] = "{}" \n'.format(experiment_token)
        required_code: str = line1 + line2 + line3 + line4 + line5 + line6

    else:
        required_code: str = line4 + line5 + line6

    click.secho("The following code snippet needs to be added after importing mlflow", fg='green')
    click.secho(required_code)
    pyperclip.copy(required_code)
    click.secho("It has also been copied to your clipboard", fg='green')



import configparser
import pathlib
import click

from deploifai.cli.utilities.config.find_config_filepath import find_config_absolute_path

"""
Manages a .dataset.cfg file to store configuration info about a project.

config dictionary structure:
{
    "DATASET": {"id": string}
}
"""

relative_config_filepath = pathlib.Path("dataset.cfg")

config_file_path = find_config_absolute_path(relative_config_filepath)

config_sections = ["DATASET"]


class DeploifaiDataAlreadyInitialisedError(Exception):
    """
    Exception when a data directory has been initialised through Deploifai in an ML project.
    """

    def __init__(self, message):
        super(DeploifaiDataAlreadyInitialisedError, self).__init__(message)


class DatasetNotInitialisedError(Exception):
    """
    Exception when dataset config is not found.
    """

    def __init__(self, message):
        super(DatasetNotInitialisedError, self).__init__(message)


def create_config_files(new_dataset_dir: str = None):
    """
    Creates the .dataset config files.
    :return: None
    """
    global config_file_path

    if new_dataset_dir is None:
        dataset_config_dir = pathlib.Path()
    else:
        dataset_config_dir = pathlib.Path(new_dataset_dir)

    config_file_path = dataset_config_dir.joinpath("dataset.cfg")

    config_file_path.touch(exist_ok=False)

    config = configparser.ConfigParser()

    with config_file_path.open("w") as config_file:
        config.write(config_file)


def read_config_file() -> configparser.ConfigParser:
    """
    Read the dataset.cfg file
    :return: A ConfigParser that contains all the configs in the config file
    """
    if config_file_path is None:
        return None

    try:
        config = configparser.ConfigParser()
        # read the config file
        config.read(config_file_path)

        for section in config_sections:
            if section not in config.sections():
                config[section] = {}

        return config
    except FileNotFoundError as err:
        click.secho("Missing .dataset.cfg file", fg="red")
        raise err


def save_config_file(config: configparser.ConfigParser):
    """
    Save the dataset config from the context in the dataset.cfg file.
    :param config: A ConfigParser object representing config data
    """
    with config_file_path.open("w") as config_file:
        config.write(config_file)


def find_config_directory() -> pathlib.Path:
    """
    Find the absolute path to the directory that contains the dataset.cfg file
    """
    return config_file_path.parent


def add_data_storage_config(data_storage_id: str, config: configparser.ConfigParser):
    """
    Add the storage configs in the existing config file. Usually run when the user initialises the directory as a data
    directory.
    :param data_storage_id: The id of the data storage from the database.
    :param config: The current config
    :return:
    """
    if (
            config is not None
            and "DATASET" in config
            and "id" in config["DATASET"]
            and len(config["DATASET"]["id"]) != 0
    ):
        raise DeploifaiDataAlreadyInitialisedError(
            "This project already has a dataset attached to it."
        )

    elif (
            config is not None
            and "DATASET" in config
    ):

        config["DATASET"] = {"id": data_storage_id}

    save_config_file(config)



import click

from .create import create
from .list import list_server
from .start import start
from .stop import stop


@click.group()
def server():
    """
    Manage training servers on Deploifai
    """
    pass


server.add_command(create)
server.add_command(list_server)
server.add_command(start)
server.add_command(stop)



import click

from .list import list_workspace
from .set import set_workspace


@click.group()
def workspace():
    """
    Manage, and browse available workspaces
    """
    pass


workspace.add_command(list_workspace)
workspace.add_command(set_workspace)



import click

from deploifai.cli.api import DeploifaiAPIError
from deploifai.cli.context import pass_deploifai_context_obj, DeploifaiContextObj, is_authenticated


@click.command("list")
@click.option("--project", "-p", help="Project Name")
@pass_deploifai_context_obj
@is_authenticated
def list_server(context: DeploifaiContextObj, project: str):
    """
    List all training servers in current workspace
    """

    current_workspace = context.global_config["WORKSPACE"]["username"]

    click.secho("Workspace Name: {}".format(current_workspace), fg="blue")

    if project:
        # checking if project exists
        fragment = """
                    fragment project on Project {
                        id 
                    }
                    """
        where_project = {"name": {"equals": project}}
        try:
            projects_data = context.api.get_projects(workspace=current_workspace,
                                                     fragment=fragment, where_project=where_project)
        except DeploifaiAPIError:
            click.secho(
                f"Project \"{project}\" cannot be found in workspace \"{current_workspace}\"",
                fg="red",
            )
            return
        if len(projects_data) == 0:
            click.secho(
                f"Project \"{project}\" cannot be found in workspace \"{current_workspace}\"",
                fg="red",
            )
            return
        click.secho("Project Name: {}".format(project), fg="blue")
        project_id = projects_data[0]["id"]
        where_training = {"project": {"is": {"id": {"equals": project_id}}}}

    elif context.local_config and "id" in context.local_config["PROJECT"]:
        project_id = context.local_config["PROJECT"]["id"]
        # query for workspace name from api
        fragment = """
                        fragment project on Project {
                            name
                        }
                        """
        context.debug_msg(project_id)
        project_data = context.api.get_project(
            project_id=project_id, fragment=fragment
        )
        project = project_data["name"]
        click.secho("Project Name: {}".format(project), fg="blue")
        where_training = {"project": {"is": {"id": {"equals": project_id}}}}

    else:
        where_training = None

    server_info = context.api.get_training_servers(current_workspace, where_training)

    if len(server_info) == 0:
        click.secho("No training servers exist", fg="yellow")
        return

    click.secho("All training servers:", fg="blue")
    for info in server_info:
        click.echo(f"{info['name']} <{info['status']}> - {info['state']}")
    return



import click
import click_spinner
from InquirerPy import prompt
from click import Abort

from deploifai.cli.api import DeploifaiAPIError
from deploifai.cli.context import (
    pass_deploifai_context_obj,
    DeploifaiContextObj,
    is_authenticated,
)
from deploifai.cli.utilities.config import local_config


@click.command()
@pass_deploifai_context_obj
@is_authenticated
def init(context: DeploifaiContextObj):
    """
    Initialize current working directory as a Deploifai project
    """
    click.secho("Connecting with an existing project", fg="blue")
    deploifai_api = context.api

    command_workspace = context.global_config["WORKSPACE"]["username"]

    if context.local_config is not None:
        click.echo("Project already initialized")
        raise click.Abort()

    fragment = """
                    fragment project on Project {
                        id
                        name
                        cloudProfile{
                            provider
                            }
                    }
                    """

    try:
        with click_spinner.spinner():
            click.echo("Getting project information")
            project_info = deploifai_api.get_projects(
                workspace=command_workspace, fragment=fragment
            )
        if not len(project_info):
            click.echo("No projects in the workspace")
            raise click.Abort()
        questions = [
            {
                "type": "list",
                "name": "project_option",
                "message": "Choose a project to initialize.",
                "choices": [
                    {
                        "name": "{} <{}>".format(
                            x["name"], x["cloudProfile"]["provider"]
                        ),
                        "value": x,
                    }
                    for x in project_info
                ],
            }
        ]
        answers = prompt(questions=questions)
        if answers == {}:
            raise Abort()
        project_info = answers["project_option"]
        context.debug_msg(project_info)
        project_id = project_info["id"]
        project_name = project_info["name"]
    except DeploifaiAPIError as err:
        click.echo(err)
        raise Abort()

    local_config.create_config_files()
    context.local_config = local_config.read_config_file()
    context.debug_msg(context.local_config)

    # storing the project information on the local.cfg file
    local_config.set_project_config(project_id, context.local_config)



from enum import Enum

class Provider(Enum):
    AWS = "AWS"
    AZURE = "AZURE"
    GCP = "GCP"

class CloudProfile:
    def __init__(self, id, name, provider, workspace):
        self.id = id
        self.name = name
        self.provider = provider
        self.workspace = workspace



import click

from deploifai.cli.context import (
    pass_deploifai_context_obj,
    DeploifaiContextObj,
    is_authenticated,
    dataset_found
)

from deploifai.cli.utilities.frontend_routing import get_dataset_route


@click.command()
@pass_deploifai_context_obj
@is_authenticated
@dataset_found
def info(context: DeploifaiContextObj):
    """
    Get information about the initialized dataset
    """
    deploifai_api = context.api

    dataset_id = context.dataset_config["DATASET"]["id"]

    storage_details = deploifai_api.get_data_storage_info(dataset_id)

    project_name = storage_details["projects"][0]["name"]
    workspace_name = storage_details["account"]["username"]
    cloud_provider = storage_details["cloudProviderYodaConfig"]["provider"]
    container = storage_details["containers"][0]
    container_name = container["cloudName"]

    click.secho("Workspace Name: {}".format(workspace_name), fg="blue")
    click.secho("Project Name: {}".format(project_name), fg="blue")
    click.secho("Dataset Name: {}".format(storage_details["name"]), fg="blue")
    click.secho("Dataset Status: {}".format(storage_details["status"]), fg="blue")

    url = get_dataset_route(workspace_name, project_name, dataset_id)
    click.secho("\nBrowse the dataset on Deploifai: \n{}\n".format(url))

    click.secho("To find this dataset on {} cloud service portal:".format(cloud_provider))

    if cloud_provider == "AZURE":
        storage_account_name = storage_details["cloudProviderYodaConfig"]["azureConfig"]["storageAccount"]
        click.secho("Storage Account Name: {}".format(storage_account_name))
    click.secho("Container Name: {}".format(container_name))



import pathlib


def find_config_absolute_path(relative_filepath: pathlib.Path) -> pathlib.Path:
    """
    Traverse up the file system and checks if a file at the relative filepath exists.
    If it does not exist, return None.
    :return: if local config file exists, return pathlib.Path object that points to the config file
    """
    path = pathlib.Path.cwd()
    while not path.joinpath(relative_filepath).exists() and path != path.parent:
        path = path.parent

    if path == path.parent:
        return None

    return path.joinpath(relative_filepath)



import typing
from json import JSONDecodeError

import requests

from deploifai.cli.api.errors import DeploifaiAPIError
from deploifai.cli.utilities.cloud_profile import CloudProfile
from deploifai.cli.utilities import environment


def _parse_cloud_profile(cloud_profile_json, workspace) -> CloudProfile:
    """
    Re-usable function that can help parse cloud profile JSON.

    For example:
    cloudProfiles{
      id
      name
      provider
    }
    :param cloud_profile_json: JSON of Cloud profile
    :param workspace: The workspace/account where the cloud profile should belong to
    :return: CloudProfile
    """
    cloud_profile = CloudProfile(
        id=cloud_profile_json["id"],
        name=cloud_profile_json["name"],
        provider=cloud_profile_json["provider"],
        workspace=workspace,
    )
    return cloud_profile


class DeploifaiAPI:
    def __init__(self, auth_token: str = None):
        self.uri = "{backend_url}/graphql".format(backend_url=environment.backend_url)
        self.headers = {"authorization": auth_token} if auth_token else {}

    def get_user(self):
        query = """
    query {
      me {
        id
        account {
          id
          username
          isTeam
        }
        teams {
          id
          account {
            id
            username
            isTeam
          }
        }
      }
    }
    """
        query_response = requests.post(
            self.uri, headers=self.headers, json={"query": query}
        )
        user_data = query_response.json()
        return user_data["data"]["me"]

    def get_containers(self, data_storage_id):
        query = """
    query($id:String){
      dataStorage(where:{id:$id}){
        containers{
          id
          directoryName
          cloudName
        }
      }
    }
    """
        r = requests.post(
            self.uri,
            json={"query": query, "variables": {"id": data_storage_id}},
            headers=self.headers,
        )
        try:
            data_storage_containers = r.json()["data"]["dataStorage"]["containers"]
            data_storage_containers = map(
                lambda d: {"name": d.get("directoryName"), "value": d.get("cloudName")},
                data_storage_containers,
            )
            return data_storage_containers
        except TypeError:
            return []

    def get_data_storages(self, workspace: str, where_data_storage: dict = None):
        query = """
      query($username:String $where: DataStorageWhereInput){
        dataStorages(whereAccount:{
          username: $username
        }, whereDataStorage: $where) {
          id
          name
          status
          account {
            username
          }
          cloudProfile{
            provider
          }
        }
      }
    """

        variables = {
            "username": workspace,
            "where": where_data_storage
        }

        query_json = {"query": query, "variables": variables}

        try:
            query_response = requests.post(
                self.uri, headers=self.headers, json=query_json
            )
            query_response_json = query_response.json()
            return query_response_json["data"]["dataStorages"]
        except JSONDecodeError as err:
            """
            Handle errors when the JSON is not decoded
            For Python3
            """
            raise DeploifaiAPIError("The API did not return a JSON object")
        except ValueError as err:
            """
            Handle errors when the JSON is not decoded
            For Python2
            """
            raise DeploifaiAPIError("The API did not return a JSON object")

    def get_data_storage_info(self, storage_id):
        query = """
    query($id:String){
      dataStorage(where:{id:$id}){
        name
        status
        account{
          username
        }
        projects {
          id
          name
        }
        containers{
          id
          directoryName
          cloudName
        }
        cloudProviderYodaConfig{
          provider
          awsConfig{
            awsRegion
            awsAccessKey
            awsSecretAccessKey
          }
          azureConfig{
            azureRegion
            storageAccount
            storageAccessKey
          }
          gcpConfig{
            gcpServiceAccountKey
          }
        }
      }
    }
    """
        r = requests.post(
            self.uri,
            json={"query": query, "variables": {"id": storage_id}},
            headers=self.headers,
        )
        storage_details = r.json()["data"]["dataStorage"]
        return storage_details

    def get_storage_account_access_key(self, storage_id: str):
        query = """
    query($id:String){
      dataStorage(where:{id:$id}){
        cloudProviderYodaConfig{
          azureConfig{
            storageAccount
            storageAccessKey
          }
        }
      }
    }
    """
        r = requests.post(
            self.uri,
            json={"query": query, "variables": {"id": storage_id}},
            headers=self.headers,
        )
        return r.json()["data"]["dataStorage"]["cloudProviderYodaConfig"][
            "azureConfig"
        ]["storageAccessKey"]

    def create_data_storage(
            self,
            storage_name: str,
            project_id: str,
            cloud_profile: CloudProfile,
            region=None,
    ):
        mutation = """
    mutation(
      $whereProject: ProjectWhereUniqueInput!
      $data: CreateDataStorageInput!
    ) {
      createDataStorage(whereProject: $whereProject, data: $data) {
        id
      }
    }
    """
        aws_config = None
        azure_config = None
        gcp_config = None
        # TODO: Support more regions
        if cloud_profile.provider == "AWS":
            aws_config = {"awsRegion": "us-east-1"}
        elif cloud_profile.provider == "AZURE":
            azure_config = {"azureRegion": "East US"}
        elif cloud_profile.provider == "GCP":
            gcp_config = {"gcpRegion": "us-central1", "gcpZone": "us-central1-a"}

        variables = {
            "whereProject": {"id": project_id},
            "data": {
                "name": storage_name,
                "cloudProfileId": cloud_profile.id,
                "cloudProviderYodaConfig": {
                    "awsConfig": aws_config,
                    "azureConfig": azure_config,
                    "gcpConfig": gcp_config
                },
            },
        }
        r = requests.post(
            self.uri,
            json={"query": mutation, "variables": variables},
            headers=self.headers,
        )
        create_mutation_data = r.json()

        if "errors" in create_mutation_data:
            raise DeploifaiAPIError(create_mutation_data['errors'][0]['message'])

        return create_mutation_data["data"]["createDataStorage"]

    def get_cloud_profiles(self, workspace: str) -> typing.List[CloudProfile]:
        query = """
            query($whereAccount: AccountWhereUniqueInput!){
                cloudProfiles(whereAccount: $whereAccount){
                    id 
                    name 
                    provider
                } 
            }
        """

        variables = {"whereAccount": {"username": workspace}}

        r = requests.post(self.uri, json={"query": query, "variables": variables}, headers=self.headers)
        api_data = r.json()

        cloud_profiles_data = api_data["data"]["cloudProfiles"]

        cloud_profiles = list(
            map(
                lambda x: _parse_cloud_profile(
                    x, workspace
                ),
                cloud_profiles_data,
            )
        )

        return cloud_profiles

    def create_cloud_profile(self, provider, name, credentials, workspace: str, fragment):
        mutation = """
            mutation(
              $whereAccount: AccountWhereUniqueInput!
              $data: CloudProfileCreateInput!
            ) {
              createCloudProfile(whereAccount: $whereAccount, data: $data) {
                ...cloud_profile
              }
            }
        """

        variables = {
            "whereAccount": {"username": workspace},
            "data": {
                "name": name,
                "provider": provider.value,
            },
        }

        if provider.value == "AWS":
            variables["data"]["awsCredentials"] = credentials
        elif provider.value == "AZURE":
            variables["data"]["azureCredentials"] = credentials
        elif provider.value == "GCP":
            variables["data"]["gcpCredentials"] = credentials

        try:
            r = requests.post(
                self.uri,
                json={"query": mutation + fragment, "variables": variables},
                headers=self.headers,
            )

            create_mutation_data = r.json()

            return create_mutation_data["data"]["createCloudProfile"]["id"]
        except TypeError:
            raise DeploifaiAPIError("Could not create cloud profile. Please try again.")
        except KeyError:
            raise DeploifaiAPIError("Could not create cloud profile. Please try again.")

    def get_projects(self, workspace: str, fragment: str, where_project=None):
        query = (
                """
            query($whereAccount: AccountWhereUniqueInput!, $whereProject: ProjectWhereInput) {
              projects(whereAccount: $whereAccount, whereProject: $whereProject) {
                ...project
              }
            }
            """
                + fragment
        )

        if where_project:
            variables = {
                "whereAccount": {"username": workspace},
                "whereProject": where_project,
            }
        else:
            variables = {"whereAccount": {"username": workspace}}

        try:
            r = requests.post(
                self.uri,
                json={"query": query, "variables": variables},
                headers=self.headers,
            )
            api_data = r.json()

            return api_data["data"]["projects"]
        except TypeError:
            raise DeploifaiAPIError("Could not get projects. Please try again.")
        except KeyError:
            raise DeploifaiAPIError("Could not get projects. Please try again.")

    def get_project(self, project_id: str, fragment: str):
        query = (
                """    
                query ($where: ProjectWhereUniqueInput!){
                    project(where: $where) {
                    ...project
                  }
                }
                """
                + fragment
        )

        variables = {
            "where": {"id": project_id},
        }

        try:
            r = requests.post(
                self.uri,
                json={"query": query, "variables": variables},
                headers=self.headers,
            )
            api_data = r.json()

            return api_data["data"]["project"]
        except TypeError:
            raise DeploifaiAPIError("Could not get project. Please try again.")
        except KeyError:
            raise DeploifaiAPIError("Could not get project. Please try again.")

    def create_project(self, project_name: str, cloud_profile: CloudProfile, fragment: str):
        mutation = """
    mutation(
      $whereAccount: AccountWhereUniqueInput!
      $data: CreateProjectInput!
    ) {
      createProject(whereAccount: $whereAccount, data: $data) {
        ...project
      }
    }
    """

        variables = {
            "whereAccount": {"username": cloud_profile.workspace},
            "data": {
                "name": project_name,
                "cloudProfileId": cloud_profile.id,
            },
        }

        try:
            r = requests.post(
                self.uri,
                json={"query": mutation + fragment, "variables": variables},
                headers=self.headers,
            )

            create_mutation_data = r.json()

            return create_mutation_data["data"]["createProject"]
        except TypeError:
            raise DeploifaiAPIError("Could not create project. Please try again.")
        except KeyError:
            raise DeploifaiAPIError("Could not create project. Please try again.")

    def get_training_infrastructure_plans(self, uses_gpu: bool, cloud_provider: CloudProfile):
        query = (
            """
            query($usesGpu: Boolean $whereProvider: CloudProvider){
            trainingInfrastructurePlans(usesGpu: $usesGpu provider: $whereProvider){
                plan
                config{
                  ... on AWSFalconDefaultConfig{
                    ec2InstanceType
                  }
                  ... on AzureFalconDefaultConfig{
                    vmSize
                  }
                  ... on GCPFalconDefaultConfig{
                    computeMachineType
                  }
                }
              }
            }
            """
        )
        variables = {
            "usesGpu": uses_gpu,
            "whereProvider": cloud_provider
        }

        try:
            r = requests.post(
                self.uri,
                json={"query": query, "variables": variables},
                headers=self.headers,
            )
            api_data = r.json()

            return api_data["data"]["trainingInfrastructurePlans"]
        except TypeError:
            raise DeploifaiAPIError("Could not get information. Please try again.")
        except KeyError:
            raise DeploifaiAPIError("Could not get information. Please try again.")

    def falcon_ml_config(self, variable):
        query = (
            """    
            query ($where: FalconMLConfigWhereInput)
                {
                    falconMLConfigs(where: $where){
                        id
                        pythonVersion
                        frameworkVersion
                        cudaVersion
                        cudnnVersion
                    }
                }
            """
        )
        variables = {
            "where": variable
        }

        try:
            r = requests.post(
                self.uri,
                json={"query": query, "variables": variables},
                headers=self.headers,
            )
            api_data = r.json()

            return api_data["data"]["falconMLConfigs"]
        except TypeError:
            raise DeploifaiAPIError("Could not get ML Configs. Please try again.")
        except KeyError:
            raise DeploifaiAPIError("Could not get ML Configs. Please try again.")

    def falcon_ml_config_distinct(self, variable, distinct_option: str):
        query = (
            """    
            query ($where: FalconMLConfigWhereInput $distinct: [FalconMLConfigScalarFieldEnum!])
                {
                    falconMLConfigs(where: $where distinct: $distinct){
                        id
                        pythonVersion
                        frameworkVersion
                        cudaVersion
                        cudnnVersion
                    }
                }
            """
        )
        variables = {
            "where": variable,
            "distinct": [distinct_option],
        }

        try:
            r = requests.post(
                self.uri,
                json={"query": query, "variables": variables},
                headers=self.headers,
            )
            api_data = r.json()

            return api_data["data"]["falconMLConfigs"]
        except TypeError:
            raise DeploifaiAPIError("Could not get ML Configs. Please try again.")
        except KeyError:
            raise DeploifaiAPIError("Could not get ML Configs. Please try again.")

    def create_training_server(self, name: str, project_id: str, cloud_profile_id: str,
                               falcon_plan: str, uses_gpu: bool,
                               data_storage_ids: typing.List[str] = None, falcon_ml_config_id: str = None):
        mutation = """
        mutation($whereProject: ProjectWhereUniqueInput! $data: CreateTrainingInput!){
            createTraining(whereProject: $whereProject data: $data){
            name
            status
            }
        }
        """

        variables = {
            "whereProject": {"id": project_id},
            "data": {
                "name": name,
                "cloudProfileId": cloud_profile_id,
                "cloudProviderFalconConfig": {
                    "plan": falcon_plan,
                    "usesGpu": uses_gpu,
                },
                "falconMLConfigId": falcon_ml_config_id,
                "dataStorageIds": data_storage_ids,
            },
        }

        try:
            r = requests.post(
                self.uri,
                json={"query": mutation, "variables": variables},
                headers=self.headers,
            )

            create_mutation_data = r.json()

            return create_mutation_data["data"]["createTraining"]
        except TypeError:
            raise DeploifaiAPIError("Could not create Training Server. Please try again.")
        except KeyError:
            raise DeploifaiAPIError("Could not create Training Server. Please try again.")

    def get_training_servers(self, workspace: str, where_training: dict = None):
        query = """
        query ($whereAccount: AccountWhereUniqueInput! $whereTraining: TrainingWhereInput){
          trainings(whereAccount: $whereAccount whereTraining: $whereTraining){
            id
            name
            status
            state
          }
        }
        """

        variables = {
            "whereAccount": {"username": workspace},
            "whereTraining": where_training,
        }

        try:
            r = requests.post(
                self.uri,
                json={"query": query, "variables": variables},
                headers=self.headers,
            )
            server_info = r.json()

            if "errors" in server_info:
                raise DeploifaiAPIError(server_info['errors'][0]['message'])

            server_details = server_info["data"]["trainings"]

            return server_details

        except TypeError:
            raise DeploifaiAPIError("Could not find Training Server. Please try again.")
        except KeyError:
            raise DeploifaiAPIError("Could not find Training Server. Please try again.")

    def start_training_server(self, server_id: str):
        mutation = """
        mutation($where : TrainingWhereUniqueInput!){
            startTraining(where: $where){
                status state
            }
        }
        """

        variables = {"where": {"id": server_id}}

        try:
            r = requests.post(
                self.uri,
                json={"query": mutation, "variables": variables},
                headers=self.headers,
            )

            create_mutation_data = r.json()

            return create_mutation_data["data"]["startTraining"]
        except TypeError:
            raise DeploifaiAPIError("Could not start Training Server. Please try again.")
        except KeyError:
            raise DeploifaiAPIError("Could not start Training Server. Please try again.")

    def stop_training_server(self, server_id: str):
        mutation = """
        mutation($where : TrainingWhereUniqueInput!){
            stopTraining(where: $where){
                status state
            }
        }
        """

        variables = {"where": {"id": server_id}}

        try:
            r = requests.post(
                self.uri,
                json={"query": mutation, "variables": variables},
                headers=self.headers,
            )

            create_mutation_data = r.json()

            return create_mutation_data["data"]["stopTraining"]
        except TypeError:
            raise DeploifaiAPIError("Could not stop Training Server. Please try again.")
        except KeyError:
            raise DeploifaiAPIError("Could not stop Training Server. Please try again.")

    def get_container_registries(self, username: str, cloud_profile_id: str):
        query = """
        query ($whereAccount: AccountWhereUniqueInput! $whereContainerRegistry: ContainerRegistryWhereInput){
            containerRegistries(whereAccount: $whereAccount whereContainerRegistry: $whereContainerRegistry){
                id
                name
            }
        }
        """

        variables = {
            "whereAccount": {"username": username},
            "whereContainerRegistry": {"cloudProfileId": {"equals": cloud_profile_id}},
        }

        try:
            r = requests.post(
                self.uri,
                json={"query": query, "variables": variables},
                headers=self.headers,
            )

            container_registries = r.json()
            return container_registries["data"]["containerRegistries"]
        except TypeError:
            raise DeploifaiAPIError("Could not get Container Registries. Please try again.")
        except KeyError:
            raise DeploifaiAPIError("Could not get Container Registries. Please try again.")

    def get_container_registry(self, registry_id: str):
        query = """
        query ($where: ContainerRegistryWhereUniqueInput!){
            containerRegistry(where: $where){
                id
                name
                info {
                    loginServer
                    username
                    password
                    imageUri
                }
            }
        }
        """

        variables = {"where": {"id": registry_id}}

        try:
            r = requests.post(
                self.uri,
                json={"query": query, "variables": variables},
                headers=self.headers,
            )

            container_registry = r.json()
            return container_registry["data"]["containerRegistry"]
        except TypeError:
            raise DeploifaiAPIError("Could not get Container Registry. Please try again.")
        except KeyError:
            raise DeploifaiAPIError("Could not get Container Registry. Please try again.")

    def create_container_registry(self, project_id: str, name: str, cloud_profile_id: str):
        mutation = """
        mutation($whereProject: ProjectWhereUniqueInput! $data: CreateContainerRegistryInput!){
            createContainerRegistry(whereProject: $whereProject data: $data){
                id
                name
                sluggedName
                info {
                    loginServer
                    username
                    password
                    imageUri
                }
            }
        }
        """

        variables = {
            "whereProject": {"id": project_id},
            "data": {
                "name": name,
                "cloudProfileId": cloud_profile_id,
                "cloudProviderContainerRegistryConfig": {},
            }
        }

        try:
            r = requests.post(
                self.uri,
                json={"query": mutation, "variables": variables},
                headers=self.headers,
            )

            create_mutation_data = r.json()

            return create_mutation_data["data"]["createContainerRegistry"]
        except TypeError:
            raise DeploifaiAPIError("Could not create Container Registry. Please try again.")
        except KeyError:
            raise DeploifaiAPIError("Could not create Container Registry. Please try again.")

    def get_application_infrastructure_plans(self, cloud_provider: str):
        query = """
        query ($provider: CloudProvider) {
            applicationInfrastructurePlans(provider: $provider) {
                plan
                config {
                    ... on AWSAppDefaultConfig{
                    ec2InstanceType
                    }
                    ... on AzureAppDefaultConfig{
                    cgCpu
                    }
                    ... on GCPAppDefaultConfig{
                    containerCpu
                    }
                }
            }
        }
        """

        variables = {
            "provider": cloud_provider,
        }

        app_configs = []
        try:
            r = requests.post(
                self.uri,
                json={"query": query, "variables": variables},
                headers=self.headers,
            )
            app_info = r.json()

            if "errors" in app_info:
                raise DeploifaiAPIError(app_info['errors'][0]['message'])

            app_configs = app_info["data"]["applicationInfrastructurePlans"]
        except Exception:
            pass
        finally:
            return app_configs

    def get_applications(self, workspace: str, where_application: dict = None):
        query = """
        query ($whereAccount: AccountWhereUniqueInput! $whereApplication: ApplicationWhereInput){
            applications(whereAccount: $whereAccount whereApplication: $whereApplication){
                id
                name
                status
            }
        }
        """

        variables = {
            "whereAccount": {"username": workspace},
            "whereApplication": where_application,
        }

        try:
            r = requests.post(
                self.uri,
                json={"query": query, "variables": variables},
                headers=self.headers,
            )
            app_info = r.json()

            if "errors" in app_info:
                raise DeploifaiAPIError(app_info['errors'][0]['message'])

            app_details = app_info["data"]["applications"]
        except TypeError:
            raise DeploifaiAPIError("Could not find Applications. Please try again.")
        except KeyError:
            raise DeploifaiAPIError("Could not find Applications. Please try again.")

        return app_details

    def get_application(self, application_id: str):
        query = """
        query ($where: ApplicationWhereUniqueInput!){
            application(where: $where){
                id
                name
                status
                hostname
            }
        }
        """

        variables = {"where": {"id": application_id}}

        try:
            r = requests.post(
                self.uri,
                json={"query": query, "variables": variables},
                headers=self.headers,
            )
            app_info = r.json()

            if "errors" in app_info:
                raise DeploifaiAPIError(app_info['errors'][0]['message'])

            app_details = app_info["data"]["application"]
        except TypeError:
            raise DeploifaiAPIError("Could not find Application. Please try again.")
        except KeyError:
            raise DeploifaiAPIError("Could not find Application. Please try again.")

        return app_details

    def create_application(self, project_id: str, name: str, cloud_profile_id: str, config: dict, image_uri: str, port: int, environment_variables: list):
        mutation = """
        mutation($whereProject: ProjectWhereUniqueInput! $data: CreateApplicationInput!){
            createApplication(whereProject: $whereProject data: $data){
                id
                name
            }
        }
        """

        variables = {
            "whereProject": {"id": project_id},
            "data": {
                "name": name,
                "cloudProfileId": cloud_profile_id,
                "cloudProviderAppConfig": config,
                "container": {
                    "image": image_uri,
                    "port": port,
                },
                "environmentVariables": environment_variables,
            }
        }

        try:
            r = requests.post(
                self.uri,
                json={"query": mutation, "variables": variables},
                headers=self.headers,
            )

            create_mutation_data = r.json()

            return create_mutation_data["data"]["createApplication"]
        except TypeError:
            raise DeploifaiAPIError("Could not create Application. Please try again.")
        except KeyError:
            raise DeploifaiAPIError("Could not create Application. Please try again.")

    def update_application(self):
        ...



import click
from deploifai.cli.dataset.init import init
from deploifai.cli.dataset.info import info
from deploifai.cli.dataset.push import push
from deploifai.cli.dataset.create import create
from deploifai.cli.dataset.list import list_data
from deploifai.cli.dataset.pull import pull


@click.group()
def dataset():
    """
    Manage datasets and dataset storages on Deploifai
    """
    pass


dataset.add_command(init)
dataset.add_command(push)
dataset.add_command(pull)
dataset.add_command(info)
dataset.add_command(create)
dataset.add_command(list_data)



import setuptools

with open("readme.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="deploifai",
    author="Deploifai Limited",
    description="Deploifai CLI",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/deploifai/cli",
    packages=setuptools.find_packages(),
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=[
        "click",
        "python-dotenv",
        "keyring",
        "requests",
        "click-spinner",
        "tqdm",
        "azure-identity",
        "azure-storage-blob",
        "boto3",
        "google-auth",
        "google-cloud-storage",
        "docker",
        "InquirerPy",
        "pyperclip",
        "semver",
        "cryptography",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            "deploifai = deploifai.cli:cli",
        ],
    },
)



import click
import click_spinner
from InquirerPy import prompt
from click import Abort

from deploifai.cli.api import DeploifaiAPIError
from deploifai.cli.context import (
    pass_deploifai_context_obj,
    DeploifaiContextObj,
    is_authenticated,
)
from deploifai.cli.utilities.config import dataset_config


@click.command()
@pass_deploifai_context_obj
@is_authenticated
def init(context: DeploifaiContextObj):
    """
    Initialise the current directory as a dataset
    """
    click.secho("Connecting with an existing dataset", fg="blue")
    deploifai_api = context.api

    command_workspace = context.global_config["WORKSPACE"]["username"]

    try:
        with click_spinner.spinner():
            click.echo("Getting dataset information")
            data_storages = deploifai_api.get_data_storages(
                workspace=command_workspace
            )
        if not len(data_storages):
            click.echo("No dataset in the workspace")
            raise click.Abort()
        questions = [
            {
                "type": "list",
                "name": "storage_option",
                "message": "Choose the dataset to link.",
                "choices": [
                    {
                        "name": "{} <{}>".format(
                            x["name"], x["cloudProfile"]["provider"]
                        ),
                        "value": x["id"],
                    }
                    for x in data_storages
                ],
            }
        ]
        answers = prompt(questions=questions)
        if answers == {}:
            raise Abort()
        storage_id = answers["storage_option"]

        context.debug_msg(storage_id)
    except DeploifaiAPIError as err:
        click.echo(err)
        raise Abort()

    try:
        dataset_config.create_config_files()
        context.dataset_config = dataset_config.read_config_file()
        context.debug_msg(context.dataset_config)
        click.echo(
            "Creating the dataset config file. If your dataset is outside the current working directory, "
            "move it into the the current working directory so that it can be uploaded."
        )
    except FileExistsError as err:
        click.echo("Using the existing dataset directory")

    try:
        dataset_config.add_data_storage_config(storage_id, context.dataset_config)
    except dataset_config.DeploifaiDataAlreadyInitialisedError:
        click.echo(
            """
    A different dataset is already initialised in the config file.
    Consider removing it from the config file, or removing the config file itself.
    """
        )
        raise Abort()



import abc
import os
import typing
from pathlib import Path
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor, as_completed

from deploifai.cli.utilities.config.dataset_config import find_config_directory


class DataStorageHandlerException(Exception):
    pass


class DataStorageHandlerEmptyFilesException(DataStorageHandlerException):
    def __str__(self):
        return "DataStorageHandlerEmptyFilesException: No files found in the dataset"


class DataStorageHandlerTargetNotFoundException(DataStorageHandlerException):
    def __str__(self):
        return "DataStorageHandlerTargetNotFoundException: Target not found"


class DataStorageHandler(abc.ABC):
    def __init__(self, dataset_id: str, container_cloud_name: str, client):
        self.id = dataset_id
        self.container_cloud_name = container_cloud_name
        self.client = client
        self.dataset_directory = find_config_directory()

    def push(self, target: Path):
        """
        Uploads files to the cloud in the target directory.
        :param target: The absolute path to the target file or directory to be uploaded.
        """
        self.upload_dataset(target)

    def pull(self, target: Path):
        """
        Downloads files from the cloud in the target directory, overwriting files.
        :param target: The absolute path to the target file or directory to be downloaded.
        """
        self.download_dataset(target)

    @staticmethod
    def upload_file(
        client, file_path: Path, object_key: str, container_cloud_name: str
    ):
        """
        Upload a given file to the cloud dataset.
        :param client: A client from the SDK of a specific cloud service provider.
        :param file_path: The absolute file path to upload.
        :param object_key: The key of the file object in cloud storage.
        :param container_cloud_name: The cloud name of the dataset.
        :return: None
        """
        pass

    def upload_dataset(self, target: Path):
        """
        This function helps upload a directory to the cloud.
        Uses a ThreadPoolExecutor to make uploads faster.
        :param target: The absolute path to target file or directory to be uploaded.
        :return: None
        """
        if not target.exists():
            raise DataStorageHandlerTargetNotFoundException()

        # if target is a file, set it into the files list
        if target.is_file():
            files = [target]
        else:
            # if target is a directory, upload all files in it
            directory_generator = Path(target).glob("**/*")
            files = [f for f in directory_generator if f.is_file()]

            if len(files) == 0:
                raise DataStorageHandlerEmptyFilesException()

        # begin uploading
        with tqdm(total=len(files)) as pbar:
            with ThreadPoolExecutor(max_workers=5) as ex:
                futures = [
                    ex.submit(
                        self.upload_file, self.client, file_path, str(file_path.relative_to(self.dataset_directory).as_posix()), self.container_cloud_name
                    )
                    for file_path in files
                ]
                for future in as_completed(futures):
                    future.result()
                    pbar.update(1)

    def list_files(self, prefix: str = None) -> typing.Generator:
        """
        Returns a generator that lists all files in a cloud dataset based on a given prefix.
        :param prefix: The prefix of the files to be listed.
        """
        pass

    @staticmethod
    def download_file(
            client, file, dataset_directory: Path, container_cloud_name: str
    ):
        """
        Download a given file from the cloud to the local dataset.
        :param client: A client from the SDK of a specific cloud service provider.
        :param file: A file object yielded by the list_files generator.
        :param dataset_directory: The absolute directory path of the local dataset.
        :param container_cloud_name: The cloud name of the dataset.
        """
        pass

    @staticmethod
    def make_dirs(object_key: str, dataset_directory: Path):
        """
        Creates the directories of the file in the local dataset.
        :param object_key: The key of the file object.
        :param dataset_directory: The absolute directory path of the local dataset.
        """
        relative_path = object_key.rsplit("/", 1)[0]
        abs_path = str(dataset_directory.joinpath(relative_path))
        os.makedirs(abs_path, exist_ok=True)

    def download_dataset(self, target: Path):
        """
        Downloads files from the cloud in the target directory.
        :param target: The absolute path to the target file or directory to be downloaded.
        """
        if target == self.dataset_directory:
            prefix = None
        else:
            prefix = target.relative_to(self.dataset_directory).as_posix()

        files = self.list_files(prefix)

        with ThreadPoolExecutor(max_workers=5) as ex:
            futures = [
                ex.submit(
                    self.download_file, self.client, file, self.dataset_directory, self.container_cloud_name
                )
                for file in files
            ]

            if len(futures) == 0:
                raise DataStorageHandlerEmptyFilesException()

            with tqdm(total=len(futures)) as pbar:
                for future in as_completed(futures):
                    future.result()
                    pbar.update(1)



import click
import webbrowser
import requests

from deploifai.cli.api import DeploifaiAPIError
from deploifai.cli.context import (
    pass_deploifai_context_obj,
    DeploifaiContextObj,
    project_found,
)
from deploifai.cli.utilities.frontend_routing import get_project_route


@click.command()
@pass_deploifai_context_obj
@project_found
@click.option("--workspace", '-w', help="Workspace name", type=str)
@click.option("--project", '-p', help="Project name", type=str)
def browse(context: DeploifaiContextObj, project: str, workspace="unassigned"):
    """
    Command to open a project in the Web Browser
    """

    if workspace:
        if project:
            # query for list of projects to obtain project id
            fragment = """
                            fragment project on Project {
                                id 
                            }
                            """
            where_project = {"name": {"equals": project}}
            try:
                projects_data = context.api.get_projects(workspace=workspace,
                                                         fragment=fragment, where_project=where_project)
            except DeploifaiAPIError:
                click.secho(
                    f"Project \"{project}\" cannot be found in workspace \"{workspace}\"",
                    fg="red",
                )
                return
            if len(projects_data) == 0:
                click.secho(
                    f"Project \"{project}\" cannot be found in workspace \"{workspace}\"",
                    fg="red",
                )
                return
            else:
                project_id = projects_data[0]["id"]

        else:
            click.secho("Missing project name", fg="yellow")
            click.secho("use --project option to pass the name of the project", fg="yellow")
            return

    elif project:
        click.secho("Missing workspace name", fg="yellow")
        click.secho("Use --workspace option to pass the name of the workspace", fg="yellow")
        return

    else:
        # assume that the user should be in a project directory, that contains local configuration file

        if "id" not in context.local_config["PROJECT"]:
            click.secho("Missing workspace name and project name", fg='yellow')
            click.secho("Use --workspace option to pass the name of the workspace, and --project option to pass the "
                        "name of the project", fg="yellow")
            return

        else:
            project_id = context.local_config["PROJECT"]["id"]

            # query for workspace name from api
            fragment = """
                            fragment project on Project {
                                name
                                account{
                                    username
                                }
                            }
                            """
            context.debug_msg(project_id)
            project_data = context.api.get_project(
                project_id=project_id, fragment=fragment
            )
            project = project_data["name"]
            workspace = project_data["account"]["username"]

    # defining the url
    url = get_project_route(workspace, project)
    context.debug_msg(f"Opening url: {url}")
    # checking if the url leads to a valid webpage
    response = requests.head(url)
    if response.status_code == 200:
        # opening the url in default browser
        webbrowser.open(url, new=2)
    else:
        click.secho("Invalid workspace or project name", fg='red')



import click

from deploifai.cli.context import (
    is_authenticated,
    pass_deploifai_context_obj,
    DeploifaiContextObj,
)


@click.command()
@is_authenticated
@pass_deploifai_context_obj
def status(context: DeploifaiContextObj):
    """
        Check login status
    """

    click.secho("You are logged in", fg="green")

    workspace = context.global_config["WORKSPACE"]["username"]
    click.secho(f"You are currently in the workspace {workspace}", fg="blue")



import os
from dotenv import load_dotenv

load_dotenv()

cli_environment = os.environ.get("DEPLOIFAI_CLI_ENVIRONMENT", "production")
is_development_env = cli_environment == "development"

backend_url = os.environ.get("DEPLOIFAI_BACKEND_URL", "https://api.deploif.ai")
frontend_url = os.environ.get("DEPLOIFAI_FRONTEND_URL", "https://deploif.ai")



import keyring

SERVICE_NAME = "deploifai-cli"


def save_auth_token(username, token):
    keyring.set_password(SERVICE_NAME, username, token)


def get_auth_token(username):
    return keyring.get_password(SERVICE_NAME, username)


def delete_auth_token(username):
    keyring.delete_password(SERVICE_NAME, username)
